/****************************************************************************
*                         CLOCK.C					    *
*       Soucast projektu TUOX                                               *
*       Autori:     Jiri Horacek                L92186                      *
*                   Alan Kornuta                                            *
*                   David Stupnanek                                         *
*       Spoluautor: Jiri Stecker                L94492                      *
*                                                                           *
*                                                                           *
****************************************************************************/

/****************************************************************************
*                                                                           *
* This file contains the code and data for the clock task.  The clock task  *
* accepts these message types:                                              *
*                                                                           *
*                                                                           *
*   SET_ALARM:   a process wants to be alerted after a specified interval   *
*   SET_TIME:    a process wants to set the real time in seconds            *
*   GET_TIME:    a process wants the real time in seconds                   *
*                                                                           *
* The input message is format m6 and m2.  The parameters are as follows:    *
*                                                                           *
*     m_type    CLOCK_PROC   FUNC    NEW_TIME                               *
* ---------------------------------------------                             *
* | SET_ALARM  | proc_nr  |func ptr |  delta  |                             *
* |------------+----------+---------+---------|                             *
* | SET_TIME   |          |         | newtime |                             *
* |------------+----------+---------+---------|                             *
* | GET_TIME   |          |         |         |                             *
* |------------+----------+---------+---------|                             *
* | CLOCK_TICK |          |         |         |                             *
* ---------------------------------------------                             *
*                                                                           *
* Reply messages are of type OK, except in the case of a CLOCK_TICK, to     *
* which no reply is generated.                                              *
*                                                                           *
* When an alarm goes off, if the caller is a user process,a SIGALRM signal  *
* is sent to it.  If it is a task, a function specified by the caller will  *
* be invoked.  This function may, for example, send a message, but only if  *
* it is certain that the task will be blocked when the timer goes off.      *
*                                                                           *
****************************************************************************/

#include "clock\clock.h"
#include "COMMON\signal.h"
#include "COMMON\glo.h"
#include "kernel\makra.h"
#include "kernel\proc_pub.h"
#include "systask\systask.h"

#include "test.cnf"
#include "test\trap.h"


PRIVATE real_time realtime;         	/* ticks since boot */
PRIVATE real_time boot_time; 		/* time in seconds of system boot */
PRIVATE real_time next_alarm;           /* time of next alarm in ticks */
PRIVATE int sched_ticks = SCHED_RATE;   /* counter: when 0, call scheduler */
PRIVATE void (*watch_dog[NR_TASKS+NR_SERVS])(int);    /* watch dog functions to call */

PRIVATE message m;

/*==========================================================================*
 *                              clock_task                                  *
 *==========================================================================*/
PUBLIC void clock_task(void)
{
  /* Main program of clock task.  It corrects realtime by adding pending
   * ticks seen only by the interrupt service, then it determines which
   * of the possible calls this is by looking at 'mc.m_type'.  Then
   * it dispatches.*/

  static int sender;

  next_alarm = 0;
  init_clock();  /* initialize 8254 */

  /* Main loop of the clock task.  Get work, process it, sometimes reply:
     Cekej na zpravu
     pri zamcenem preruseni pricti lost_ticks */

  while (TRUE)
    {
      int dontReply=0;
      receive (ANY, m);
      switch (m.m_type)
	{
	  case CLOCK_TICK: do_clock_tick ();
			   dontReply=1;
			   break;

	  case GET_TIME:   do_get_time (&m);
			   break;

	  case SET_TIME:   do_set_time (&m);
			   break;

	  case SET_ALARM:  do_set_alarm (&m);
			   break;

	  default:         trap (TRAP_UNKNOWN_MSG);
	}
      sender=m.m_source;
      if (!dontReply) send(sender,m);
    };
}

/*==========================================================================*
 *                              do_set_alarm                                *
 * args:                                                                    *
 *      proc_nr                         - cislo procesu                     *
 *      func_ptr                        - ukazatel na alarm funkci (u tasku)*
 *      delta                           - za jak dlouho v sekundach se ma   *
 *					  poslat alarm  		    *
 *==========================================================================*/
PRIVATE void do_set_alarm(message* m_ptr)
{
  /* A process wants an alarm signal or a task wants a given watch_dog function
     called after a specified interval.*/
  /* Poznamenej cas do proc[].p_alarm, u tasku adresu funkce do watch_dog[]
     Nastav next_alarm v ticich */

  if (m_ptr->NEW_TIME==0)
   {
    proc[m_ptr->CLOCK_PROC].p_alarm=0;
    m_ptr->RET_ERROR_CODE = ERROR;
    m_ptr->m_type = TASK_REPLY;
    /* co next_alarm ? Snad nic, dojde-li k alarmu, ktery uz neplati,
       nic se neprovede (kdyz p_alarm=0) */
   }
  else
   {
    proc[m_ptr->CLOCK_PROC].p_alarm = m_ptr->NEW_TIME*HZ + realtime;
    if (istaskp(proc_addr(m_ptr->CLOCK_PROC)))  /* proces je task */
      {
	watch_dog [m_ptr->CLOCK_PROC] = m_ptr->FUNC;  /* polozka v watch_dog
						  odpovida polozce v proc[] */
      }
    if (next_alarm == 0 || m_ptr->NEW_TIME*HZ + realtime < next_alarm)
      {
	next_alarm = m_ptr->NEW_TIME*HZ + realtime;
      }
    m_ptr->RET_ERROR_CODE = OK;
    m_ptr->m_type = TASK_REPLY;
   }
}


/*==========================================================================*
 *                              do_get_time                                 *
 * ret:                                                                     *
 *      seconds                       - pocet sekund                        *
 *      milliseconds                  - pocet milisekund                    *
 *==========================================================================*/
PRIVATE void do_get_time(message * m_ptr)
{
  /* Get and return the current clock time in seconds and milliseconds. */

  m_ptr->RET_REAL_TIME_S = boot_time + (realtime/HZ);/* pocet sekund od bootu */
  m_ptr->RET_REAL_TIME_M = (realtime%HZ); /* pocet zbylych milisekund */
  m_ptr->RET_ERROR_CODE = OK;             /* navratovy chybovy kod */
  m_ptr->m_type = TASK_REPLY;             /* typ zpravy */
}


/*==========================================================================*
 *                              do_set_time                                 *
 * args:                                                                    *
 *      new time                        - novy cas v sekundach od 1.1.1970  *
 *==========================================================================*/
PRIVATE void do_set_time(message* m_ptr)
{
  /* Set the real time clock.*/

   boot_time = m_ptr->NEW_TIME - (boot_time + (realtime/HZ));
   m_ptr->RET_ERROR_CODE = OK;            /* navratovy chybovy kod */
   m_ptr->m_type = TASK_REPLY;            /* typ zpravy */
}


/*==========================================================================*
 *                              do_clock_tick                               *
 *==========================================================================*/
PRIVATE void do_clock_tick(void)
{
  int i;
  real_time t;
  /* Volano kazdy z PROC.C::interrupt() (idealne kazdy tik hodin,
     jinak vzdy, kdyz je clock task ochoten zpravu prevzit (tehdy se musi
     kompenzovat lost_ticks) */

  /*  k real_time pricti lost_ticks takto (kvuli race condition):
      t=lost_ticks; realtime+=t +1 (+1 je za tento tik), lost_ticks-=t;*/

  t = lost_ticks;
  realtime += t + 1;

  if ((next_alarm != 0) && (next_alarm <= realtime))
   // je treba vyridit alarmy
    {
      next_alarm = 0;
      for (i=0; i<NR_TASKS+NR_SERVS+NR_PROCS; i++)
	 /* hledam procesy k vyvolani alarmu a vyvolam jej */
	{
	  if (!(proc[i].p_flags & P_SLOT_FREE) && proc[i].p_alarm!=0 && proc[i].p_alarm<=realtime)
	    {
	      if (i < NR_TASKS+NR_SERVS) watch_dog[i] (0);
	       else cause_sig(i,SIGALRM); /* v systasku */
	      proc[i].p_alarm = 0;
	    }
	  else
	    {                   /* hledam nejmensi next_alarm */
	      if ((proc[i].p_alarm != 0) && !(proc[i].p_flags & P_SLOT_FREE))
		{
		  if (next_alarm == 0) next_alarm = proc[i].p_alarm;
		  else
		    if (proc[i].p_alarm < next_alarm)
			next_alarm = proc[i].p_alarm;
		}
	    }
	}
    }

  accounting ();

  /* Zmensi sched_ticks */
  sched_ticks -= lost_ticks + 1;
  lost_ticks -= t;

  if (sched_ticks <= 0 && bill_ptr == prev_ptr)
    {
      sched ();
      sched_ticks = SCHED_RATE;
    }
  /*  je-li 0 a bezi_li user process (bill_ptr==prev_ptr):
  preplanuj sched(), reset kvantum */

  /* (prev_ptr je proces, ktery bezel pred vstupem do CLOCK tasku,
  protoze v proc_ptr je zde vzdy CLOCK) */
}

/*==========================================================================*
 *                                accounting                                *
 *==========================================================================*/
PRIVATE void accounting(void)
{
   /* bill_ptr vzdy ukazuje na proces, kteremu se uctuje za cas CPU.
      Podle toho, v jakem kodu byl CPU pred tikem (prev_proc), pricti k
      user_time nebo system_time */

  if (isuserp(prev_ptr)) /* proces je user process */
      bill_ptr->user_time += 1 + lost_ticks;
  else
      bill_ptr->sys_time += 1 + lost_ticks;
}


//Pomocne funkce:

/*==========================================================================*
 *                              init_clock                                  *
 *==========================================================================*/
PRIVATE void init_clock(void)
{
 const puv_rok = 1970;

 static int dny[2][13] = {{0,31,28,31,30,31,30,31,31,30,31,30,31},
			 {0,31,29,31,30,31,30,31,31,30,31,30,31}};

 char pom[2];
 static unsigned int rok=0,mesic=0,den=0,hodina=0,minuta=0,sekunda=0;
 unsigned long pocet=0;
 int i,prestupny,pom_rok;

  /* Initialize channel 0 of the 8253A timer to 60 Hz. */

  /* 1193180 / 60 = 19886 = 4DAEh */
  asm {
	mov al, 36h     /* (00110110) ... mod 3, 16b citac, cteni resp. zapis
			   obou Bytu slova, kanal 0 */
	out 43h, al     /* zapis do ridiciho registru */
	mov al, 0AEh    /* dolni slabika delici konstanty */
	out 40h, al
	mov al, 4Dh     /* horni slabika delici konstanty */
	out 40h, al
      }

  /* inicializovat boottime z CMOS: */
  asm {
	mov cx,10
	mov al,9        	/* vrat rok */
	out 70h,al
/* Snad kratke poseckani ? */
	in al,71h
	mov ah,al               /* uschovani roku */
	shr ah,1
	shr ah,1
	shr ah,1
	shr ah,1	/* posun desitek dolu */
	and ax,0f0fh
	xchg ah,al
	mov bx,ax
	xor ah,ah
	mov cx,10
	mul cx
	add al,bh               /* prevod na cislo */
	mov bx,offset rok
	mov [bx],al		/* zapis do promenne */
      }

  if (rok<97) rok+=2000; /* Rok vzniku TUOXe je 1997 :-) */
  else rok+=1900;
  asm {
	mov al,8		/* vrat mesic */
	out 70h,al
/* Snad kratke poseckani ? */
	in al,71h
	mov ah,al
	shr ah,1
	shr ah,1
	shr ah,1
	shr ah,1

	and ax,0f0fh
	xchg ah,al
	mov bx,ax
	xor ah,ah
	mov cx,10
	mul cx
	add al,bh
	mov bx,offset mesic
	mov [bx],al

	mov al,7		/* vrat den */
	out 70h,al
	in al,71h
	mov ah,al
	shr ah,1
	shr ah,1
	shr ah,1
	shr ah,1

	and ax,0f0fh
	xchg ah,al
	mov bx,ax
	xor ah,ah
	mov cx,10
	mul cx
	add al,bh
	mov bx,offset den
	mov [bx],al

	mov al,4		/* vrat hodinu */
	out 70h,al
/* Snad kratke poseckani ? */
	in al,71h
	mov ah,al
	shr ah,1
	shr ah,1
	shr ah,1
	shr ah,1

	and ax,0f0fh
	xchg ah,al
	mov bx,ax
	xor ah,ah
	mov cx,10
	mul cx
	add al,bh
	mov bx,offset hodina
	mov [bx],al

	mov al,2		/* vrat minutu */
	out 70h,al
/* Snad kratke poseckani ? */
	in al,71h
	mov ah,al
	shr ah,1
	shr ah,1
	shr ah,1
	shr ah,1

	and ax,0f0fh
	xchg ah,al
	mov bx,ax
	xor ah,ah
	mov cx,10
	mul cx
	add al,bh
	mov bx,offset minuta
	mov [bx],al

	mov al,0		/* vrat sekundu */
	out 70h,al
/* Snad kratke poseckani ? */
	in al,71h
	mov ah,al
	shr ah,1
	shr ah,1
	shr ah,1
	shr ah,1

	and ax,0f0fh
	xchg ah,al
	mov bx,ax
	xor ah,ah
	mov cx,10
	mul cx
	add al,bh
	mov bx,offset sekunda
	mov [bx],al
      }

  pom_rok=puv_rok;
  for (i=0;i<(rok-puv_rok);i++)
   { 					/* spocteni poctu prechodnych roku */
    if (pom_rok%4==0 && pom_rok%100!=0 || pom_rok%400==0)
     {					/* rok neni prestupny: */
      pocet+=365;
      pom_rok++;
     }
    else
     {					/* rok je prestupny: */
      pocet+=366;
      pom_rok++;
     }
   }
  pocet=pocet*24*60*60;			/* prepocet na sekundy */
  prestupny=rok%4==0 && rok%100!=0 || rok%400==0;
  for (i=1;i<mesic;i++) den+=dny[prestupny][i];
  pocet+=den*24+hodina*60+minuta*60+sekunda; /* pricteni poctu sekund od
						1.1.tohoto roku */
  boot_time=pocet;
  realtime=0; // P.G.
}


/*===========================================================================*
 *                              clock_stop                                   *
 *===========================================================================*/
PUBLIC void clock_stop(void)
{
  /* Reset the clock to the BIOS rate. (For rebooting) */

  asm {
	mov al, 36h     /* (00110110) ... mod 3, 16b citac, cteni resp. zapis
			   obou Bytu slova, kanal 0 */
	out 43h, al     /* zapis do ridiciho registru */
	mov al, 0FFh    /* dolni slabika delici konstanty */
	out 40h, al
	mov al, 0FFh    /* horni slabika delici konstanty */
	out 40h, al
      }
}

