/* Buffer (block) cache.  To acquire a block, the requesting routine calls
 * get_block(), telling which block it wants.  The block is then regarded
 * as "in use" and has its 'b_count' field incremented. 
 * All the blocks that are not in use are chained together in an LRU list, 
 * with 'front' pointing  to the least recently used block, and 'rear'  
 * to the most recently used block.  
 * A reverse chain, using the field b_prev is also maintained.
 * Usage for LRU is measured by the time the put_block() is done.  The second
 * parameter to put_block() can violate the LRU order and put a block on the
 * front of the list, if it will probably not be needed soon.  If a block
 * is modified, the modifying routine must set b_dirt to DIRTY, so the block
 * will eventually be rewritten to the disk.
 */
#ifndef __BUF_H__
#define __BUF_H__


#include "common/type.h"
#include "common/const.h"
#include "fs/common/fsconst.h"
#include "fs/common/fstype.h"

#ifndef EXTERN
#define EXTERN extern
#define EXTERN_OVERRIDE
#endif

struct dir_struct {
 inode_t d_inum;
 char d_name[NAME_SIZE];
};

/* Buffer struct - consists of header and a frame for data */

EXTERN struct buf {
  /* Data portion of the buffer. Can be treated as various formats */
  /* or uninterpreted */
  union {
    char b__data[BLOCK_SIZE];                /* ordinary user data */
    struct dir_struct b__dir[NR_DIR_ENTRIES];    /* directory block */
    zone_t b__ind[NR_INDIRECTS];             /* indirect block */
    d_inode b__inode[INODES_PER_BLOCK];      /* inode block */
  } b;

  /* Header portion of the buffer. */
  struct buf *b_next;           /* used to link all free bufs in a chain */
  struct buf *b_prev;           /* used to link all free bufs the other way */
  struct buf *b_hash;           /* used to link bufs on hash chains */
  block_t b_blocknr;            /* block number */
  dev_t b_dev;                  /* major | minor device where block resides */
  char b_dirt;                  /* CLEAN or DIRTY */
  char b_count;                 /* number of users of this buffer */
} buf[NR_BUFS];

/* A block is free if b_dev == NO_DEV. */

#define NIL_BUF ((struct buf *) 0)      /* indicates absence of a buffer */


EXTERN struct buf *buf_hash[NR_BUF_HASH];       /* the buffer hash table */

EXTERN struct buf *front;       /* points to least recently used free block */
EXTERN struct buf *rear;        /* points to most recently used free block */
EXTERN int bufs_in_use;         /* # bufs currently in use (not on free list)*/

/* When a block is released, the type of usage is passed to put_block(). */
/* From this, it is decided where to put block in LRU chain and whether */
/* it has to be written to disk immediately */

#define WRITE_IMMED        0100 /* block should be written to disk now */
#define ONE_SHOT           0200 /* set if block not likely to be needed soon */

#define INODE_BLOCK       (0 | WRITE_IMMED)        /* inode block */
#define DIRECTORY_BLOCK   (1 | WRITE_IMMED)        /* directory block */
#define INDIRECT_BLOCK    (2 | WRITE_IMMED)        /* pointer block */
#define MAP_BLOCK         (3 | WRITE_IMMED)        /* bit map */
#define ZUPER_BLOCK       (4 | WRITE_IMMED | ONE_SHOT)   /* super block */
#define FULL_DATA_BLOCK    5                             /* data, fully used */
#define PARTIAL_DATA_BLOCK 6                             /* data, partly used*/

#define HASH_MASK (NR_BUF_HASH - 1)     /* mask for hashing block numbers */

#ifdef EXTERN_OVERRIDE
#undef EXTERN
#undef EXTERN_OVERRIDE
#endif


#endif
