/*************************************************************************
 * 			inode.c - implementation file			 *
 * 		   A part of TUOX file system project '97                *
 *			  by Dalibor Lukas, L 94578                      *
 *  		    who made some corrections in a source		 *
 *		      from Petr Kolonicny .. TUOX FS '96		 *
 *************************************************************************

 * This file manages the inode table.  There are procedures to allocate and
 * deallocate inodes, acquire, erase, and release them, and read and write
 * them from the disk.
 *
 * The entry points into this file are
 *   get_inode:    search inode table for a given inode; if not there, read it
 *   put_inode:    indicate that an inode is no longer needed in memory
 *   alloc_inode:  allocate a new, unused inode
 *   free_inode:   mark an inode as available for a new file
 *   dup_inode:    indicate that someone else is using an inode table entry
 *
 *   wipe_inode:   erase some fields of a newly allocated inode
 *   rw_inode:     read a disk block and extract an inode, or corresp. write
 */

#include "lib/include/string.h"

#define EXTERN extern

#include "fs/cache/buf.h"
#include "fs/inode/super.h"

#include "fs/common/fproc.h"
#include "fs/common/file.h"
#include "fs/common/fsglo.h"

#include "kernel/makra.h"
#include "clock/h_clock.h"

#undef EXTERN

#include "fs/inode/inode.h"

//struct fproc_struct *fp;

/* struct inode inodes[NR_INODES];	see "inode.h" */

/*===========================================================================*
 *                              get_inode                                    *
 *===========================================================================*/
PUBLIC struct inode *get_inode(dev_t device, inode_t inode)
{/* Find a slot in the inode table, load the specified inode into it, and
  * return a pointer to the slot.  If 'dev' == NO_DEV, just return a free
  * slot. */

  struct inode *pinode;
  register int i;

  if (device != NO_DEV)
  { /* find out if the specified inode is already in the inode table */
    for (i = 0; i < NR_INODES; i++) {
      pinode = inodes+i;
      if (pinode->i_dev == device && pinode->i_num == inode) {
	pinode->i_count++;
	return pinode;
      }
    }
  }

  for (i = 0; i < NR_INODES; i++)
  { /* find a free slot in the inode table */
    pinode = inodes+i;
    if (pinode->i_count == 0) {
      if (device == NO_DEV) return pinode;
      pinode->i_dev = device;
      pinode->i_num = inode;
      pinode->i_count = 1;

      rw_inode(pinode,READING);
      return pinode;
    }
  }

  return NIL_INODE;
}


/*===========================================================================*
 *                              put_inode                                    *
 *===========================================================================*/
PUBLIC void put_inode(register struct inode *rip)
{ /* The caller is no longer using this inode.  If no one else is using it
   * either write it back to the disk immediately.  If it has no links,
   * truncate it and return it to the pool of available inodes.
   */

  struct buf *pIndirectZone1,*pIndirectZone2;
  register int i,j;

  if (--rip->i_count == 0) { /* i_count == 0 means no one is using it now */
    if (rip->i_nlinks == 0) { /* release blocks occupated by the inode */

      for (i = 0; i < NR_DZONE_NUM; i++) /* direct access to block numbers */
	if (rip->i_zone[i] != NO_ZONE)
	  free_zone(rip->i_dev,rip->i_zone[i]);

      if (rip->i_zone[NR_DZONE_NUM] != NO_ZONE)
      { /* indirect access to zone numbers */
	pIndirectZone1 = get_block(rip->i_dev,rip->i_zone[NR_DZONE_NUM],
		NORMAL);
	for (i = 0; i < NR_INDIRECTS; i++)
	  if (pIndirectZone1->b.b__ind[i] != NO_ZONE)
	    free_zone(rip->i_dev,pIndirectZone1->b.b__ind[i]);
	free_zone(rip->i_dev,rip->i_zone[NR_DZONE_NUM]);
      }

      if (rip->i_zone[NR_DZONE_NUM+1] != NO_ZONE)
      {	/* double indirect access to zone numbers */
	pIndirectZone1 = get_block(rip->i_dev,rip->i_zone[NR_DZONE_NUM+1],
				   NORMAL);
	for (i = 0; i < NR_INDIRECTS; i++)
	  if (pIndirectZone1->b.b__ind[i] != NO_ZONE) {
	    pIndirectZone2 = get_block(rip->i_dev,pIndirectZone1->b.b__ind[i],
				       NORMAL);
	    for (j = 0; j < NR_INDIRECTS; j++)
	      if (pIndirectZone2->b.b__ind[j] != NO_ZONE)
		free_zone(rip->i_dev,pIndirectZone2->b.b__ind[j]);
	    free_zone(rip->i_dev,pIndirectZone1->b.b__ind[i]);
	  }
	free_zone(rip->i_dev,rip->i_zone[NR_DZONE_NUM+1]);
      }

      free_inode(rip->i_dev,rip->i_num);
      wipe_inode(rip);
    }
    if (rip->i_dirt == DIRTY) rw_inode(rip, WRITING);
  }
}

/*===========================================================================*
 *                              alloc_inode                                  *
 *===========================================================================*/
PUBLIC struct inode *alloc_inode(dev_t device, mode_t bits)
{ /* Allocate a free inode on 'device', and return a pointer to it. */

  struct super_block *pSuperblock;
  bit_t bitNumber;
  struct inode *pInode;

  /* allocate bit in the superblock inode bitmap */
  pSuperblock = get_super(device);
  if ((bitNumber = alloc_bit(pSuperblock,INODE_BITMAP)) == 0)
    return NIL_INODE;

  /* load inode */
  if ((pInode = get_inode(device,bitNumber)) == NIL_INODE) {
    free_bit(pSuperblock,INODE_BITMAP,bitNumber);
    return NIL_INODE;
  }
  pInode->i_mode = bits;
  pInode->i_nlinks = 1;
  pInode->i_uid = fp->fp_realuid;	 /*pInode->i_uid = 0;*/
  pInode->i_gid = fp->fp_realgid; 	 /*pInode->i_gid = 0;*/
  pInode->i_dev = device;
  pInode->i_count = 1;
  wipe_inode(pInode);
  return pInode;
}


/*===========================================================================*
 *                              wipe_inode                                   *
 *===========================================================================*/
PUBLIC void wipe_inode(register struct inode *rip)
{ /* Erase some fields in the inode.  This function is called from
   * alloc_inode() when a new inode is to be allocated, and from truncate(),
   * when an existing inode is to be truncated. */

  register int i;
  static message msg;

  rip->i_size = 0;
  rip->i_dirt = DIRTY;
  msg.m_source = FS;
  msg.m_type = GET_TIME;
  send_recv(CLOCK,msg);
  rip->i_mtime = msg.RET_REAL_TIME_S;
  for (i = 0; i < NR_ZONE_NUMS; i++)
    rip->i_zone[i] = NO_ZONE;
}


/*===========================================================================*
 *                              free_inode                                   *
 *===========================================================================*/
PUBLIC void free_inode(dev_t device, inode_t inode)
{ /* Return an inode to the pool of unallocated inodes. */

  struct super_block *pSuperblock = get_super(device);

  free_bit(pSuperblock,INODE_BITMAP,inode);
}

/*===========================================================================*
 *                              rw_inode                                     *
 *===========================================================================*/
PUBLIC void rw_inode(register struct inode *rip, int rw_flag)
{ /* An entry in the inode table is to be copied to or from the disk. */

  struct super_block *pSuperblock = get_super(rip->i_dev);
  block_t blockNumber = 1 /*boot block*/ + 1 /*super block*/ +
	pSuperblock->s_imap_blocks + pSuperblock->s_zmap_blocks +
	rip->i_num/INODES_PER_BLOCK;
  struct buf *pBlock = get_block(rip->i_dev,blockNumber,NORMAL);

  if (rw_flag == READING)
    memcpy(rip,pBlock->b.b__inode + (rip->i_num-1)%INODES_PER_BLOCK,
	   sizeof(d_inode));
  else if (rw_flag == WRITING) {
    memcpy(pBlock->b.b__inode + (rip->i_num-1)%INODES_PER_BLOCK,rip,
	   sizeof(d_inode));
    pBlock->b_dirt = DIRTY;
    put_block(pBlock,INODE_BLOCK);
  }
}

/*===========================================================================*
 *                              dup_inode                                    *
 *===========================================================================*/
PUBLIC void dup_inode(struct inode *ip)
{ /* This routine is a simplified form of get_inode() for the case where
   * the inode pointer is already known and anyone else wants to use inode. */

  ip->i_count++;
}


/*===========================================================================*
 *                              init_inodes                                  *
 *===========================================================================*/

PUBLIC void init_inodes(void) {
 int i;
 for (i=0; i<NR_INODES; i++)
  { inodes[i].i_dev=NO_DEV; inodes[i].i_count=NO_DEV; }
}