/* inode.h - header file
 *
 * Inode (memory) table.  This table holds inodes that are currently in use.
 * In some cases they have been opened by an open() or creat() system call,
 * in other cases the file system itself needs the inode for some reason,
 * such as to search a directory for a path name.
 * The first part of the struct holds fields that are present on the
 * disk; the second part holds fields not present on the disk.
 */

#ifndef __inode_h
#define __inode_h

#include "common/const.h"
#include "common/type.h"
#include "common/com.h"
#include "fs/cache/cache.h"

#include "fs/common/fsconst.h" 


EXTERN struct inode {
	mode_t i_mode;                /* file type, protection, etc. */
	uid_t i_uid;                  /* user id of the file's owner */
	off_t i_size;                 /* current file size in bytes */
	real_time i_mtime;            /* when was file data last changed */
	gid_t i_gid;                  /* group number */
	link_t i_nlinks;              /* how many links to this file */
	zone_t i_zone[NR_ZONE_NUMS];  /* zone numbers for direct, ind, and dbl ind */

/* When inode describes special file, i_zone[0] specifies major & minor dev number */

	/* The following items are not present on the disk. */
	dev_t i_dev;                  /* which device is the inode on */
	inode_t i_num;                /* inode number on its (minor) device */
	int i_count;                  /* # times inode used; 0 means slot is free */
	char i_dirt;                  /* CLEAN or DIRTY */
	char i_mount;                 /* this bit is set if file mounted on */
} inodes[NR_INODES];


#define NIL_INODE (struct inode *) 0    /* indicates absence of inode slot */

/* Field values.  Note that CLEAN and DIRTY are defined in "const.h" */
#define NO_PIPE            0    /* i_pipe is NO_PIPE if inode is not a pipe */
#define I_PIPE             1    /* i_pipe is I_PIPE if inode is a pipe */
#define NO_MOUNT           0    /* i_mount is NO_MOUNT if file not mounted on*/
#define I_MOUNT            1    /* i_mount is I_MOUNT if file mounted on */


PUBLIC struct inode *get_inode(dev_t device, inode_t inode);
/* dev_t device;                device on which inode resides
 * inode_t inode;               inode number
 *
 * finds a slot in the inode table, load the specified inode into it, and
 * return a pointer to the slot.  If 'dev' == NO_DEV, just return a free slot.
 */


PUBLIC void put_inode(register struct inode *rip);
/* register struct inode *rip;     pointer to inode to be released
 *
 * The caller is no longer using this inode.  If no one else is using it either
 * write it back to the disk immediately.  If it has no links, truncate it and
 * return it to the pool of available inodes.
 */


PUBLIC struct inode *alloc_inode(dev_t device, mode_t bits);
/* dev_t device;                device on which to allocate the inode
 * mode_t bits;                 mode of the inode
 *
 * Allocate a free inode on 'device', and return a pointer to it. */


PUBLIC void wipe_inode(register struct inode *rip);
/* register struct inode *rip;     the inode to be erased
 *
 * Erase some fields in the inode.  This function is called from alloc_inode()
 * when a new inode is to be allocated, and from truncate(), when an existing
 * inode is to be truncated.
 */


PUBLIC void free_inode(dev_t device, inode_t inode);
/* dev_t device;                    on which device is the inode
 * inode_t inode;                   number of inode to be freed
 *
 * Return an inode to the pool of unallocated inodes. */


PUBLIC void rw_inode(register struct inode *rip, int rw_flag);
/* register struct inode *rip;     pointer to inode to be read/written
 * int rw_flag;                    READING or WRITING
 * An entry in the inode table is to be copied to or from the disk.*/


PUBLIC void dup_inode(struct inode *ip);
/* struct inode *ip;               The inode to be duplicated. */


PUBLIC void init_inodes(void);
/* initialization of inodes[] table */
#endif
