/*************************************************************************
 * 			read.c - implementation file			 *
 * 		   A part of TUOX file system project '97                *
 *			  by Dalibor Lukas, L 94578                      *
 *  		    who made some corrections in a source		 *
 *		      from Babiuch, Buril .. TUOX FS '96		 *
 *************************************************************************

/* This file contains the heart of the mechanism used to read (and write)
 * files.  Read and write requests are split up into chunks that do not cross
 * block boundaries.  Each chunk is then processed in turn.  Reads on special
 * files are also detected and handled.
 *
 * The entry points into this file are
 *   do_read:    perform the READ system call by calling read_write
 *   read_write: actually do the work of READ and WRITE
 *   read_map:   given an inode and file position, look up its zone number
 *   rw_user:   call the kernel to read and write user space
 */

#define EXTERN extern

#include "common/errno.h"
#include "common/com.h"
#include "common/type.h"
#include "common/stat.h"

#include "fs/common/fsconst.h"
#include "fs/common/fsglo.h"
#include "fs/common/fsmes.h"
#include "fs/common/fproc.h"
#include "fs/common/file.h"
#include "fs/inode/super.h"
#include "fs/device/device.h"

#include "kernel/makra.h"
#include "clock/h_clock.h"
#include "systask/syst_pub.h"
#include "fs/main.h"

#undef EXTERN
#define EXTERN

#include "fs/inode/read.h"

#undef EXTERN

PRIVATE int rw_chunk(register struct inode *rip, off_t position, int ofs, int nbytes,
		     int rw_flag, char *buff, int seg, int usr);

/*===========================================================================*
 *                              DO_READ                                      *
 *===========================================================================*/
PUBLIC int do_read()
{ /* Perform the READ system call by calling read_write */
  procnr_t caller, destproc;

  /* otestuj, zda nacitame pro nekoho jineho nez pro volajici proces  */
  caller=who;
  destproc=fs_msg.fs_seg >> SEG_MASK_SHIFT;
  if (destproc) who=destproc;
   /* pracujeme pro jiny proces nez pro volajiciho (C_EXEC) */

  fs_msg.fs_seg &= SEG_MASK;

  fs_repmsg.m1_i1=read_write(fs_msg.fs_fd,fs_msg.fs_buffer,
				fs_msg.fs_nbytes, READING, fs_msg.fs_seg);
  who=caller; /* obnov puvodniho volajiciho */
  return fs_repmsg.m1_i1;
}


/*===========================================================================*
 *                              DO_WRITE                                     *
 *===========================================================================*/
PUBLIC int do_write()
{ /* Perform the WRITE system call by calling read_write */

  return fs_repmsg.m1_i1=read_write(fs_msg.fs_fd, fs_msg.fs_buffer,
				fs_msg.fs_nbytes, WRITING, fs_msg.fs_seg);
}

#define MIN(a,b) ((a)<(b) ? (a) : (b))
/*===========================================================================*
 *                              read_write                                   *
 *===========================================================================*/
PUBLIC int read_write (int fd, char *buffer, int nbytes, int rw_flag, int seg)
/*int rw_flag;                     READING or WRITING */
{ /* Actually do the work of READ and WRITE */
  struct inode *pinode;
  off_t position,orig_pos;
  Word count,c;
  static message msg;
  int i, result, o;

  if (fp->fp_filp[fd] == NIL_FILP)
    return EBADF;			/* Bad file descriptor */
  if (rw_flag == READING) {
    if ((fp->fp_filp[fd]->filp_mode & S_IREAD)==0)
      return EACCES;
  }
  else if (rw_flag == WRITING) {
    if ((fp->fp_filp[fd]->filp_mode & S_IWRITE)==0)
      return EACCES;
  }
  else
    return EINVAL;

  pinode = fp->fp_filp[fd]->filp_ino;	/* Find the inode */
  position = fp->fp_filp[fd]->filp_pos; /* Find the position */

  if ((pinode->i_mode&I_TYPE) == I_CHAR_SPECIAL) /* rw within char dev. */
  {
    result = dev_io(rw_flag,pinode->i_zone[0],(block_t)fp->fp_filp[fd]->filp_pos,
		    nbytes,who,buffer);

    if (result == SUSPEND) {		/* The operation has been suspended */
      fp->fp_fd = fd;
      fp->fp_buffer = buffer;
      fp->fp_nbytes = nbytes;
      fp->fp_task = d_major(pinode->i_zone[0]);
      fp->fp_suspended = SUSPENDED;
      dont_reply = 1;
    }
    return result;
  }

  if (rw_flag == WRITING)
    if ((pinode->i_mode & I_TYPE) != I_BLOCK_SPECIAL)
     /* (character special devs solved above) */
      if (get_super(pinode->i_dev) != NIL_SUPER)
	if (get_super(pinode->i_dev)->s_max_size < position+nbytes)
	  return EFAULT;		      /* Out of file address space */

  if (rw_flag == WRITING) count = nbytes;
   else if (rw_flag == READING) {
    if (position+nbytes > pinode->i_size)
      count = pinode->i_size - position;
    else count = nbytes;
  }

  orig_pos=position;
  do {
    o=position % BLOCK_SIZE;
    c=MIN(count,BLOCK_SIZE-o);
    result = rw_chunk(pinode,position,o,c,rw_flag,buffer,seg,who);
    if (result<0) { count=result; break; }
    count-=c;
    position+=c;
    buffer+=c;
  } while (count>0);

  if (count==0) count=position-orig_pos;


  if ((pinode->i_mode&I_TYPE) != I_BLOCK_SPECIAL) {

    /* Modify file size */
    if (position > pinode->i_size)
      pinode -> i_size = position;

    /* Modify the time of the last change */
    if (rw_flag == WRITING) {
      msg.m_source = FS;
      msg.m_type = GET_TIME;
      send_recv(CLOCK,msg);
      pinode->i_mtime = msg.RET_REAL_TIME_S;
    }
    pinode->i_dirt=TRUE;
  }
  fp->fp_filp[fd]->filp_pos = position;	/* Update position in fd slot */
  return (int)count;		/* Return the number of moved bytes */

/* Overeni platnosti descriptoru */
/* Jde-li o character special file: */
/* zvetsi position */
/*Jde-li o zapis  */
/* ... a neni SPECIAL_BLOCK device: zjisti, */
/* zda muze byt soubor tak dlouhy (ze superbloku)*/

/*Proved vlastni prenos (rw_chunk(), primo do user space). Vykona se
po blocich ( s pouzitim rw_chunk() ).
Pro pripad write do normalniho souboru: modifikuj cas souboru a delku
Update pozice po operaci ve filp[]*/
}


/*===========================================================================*
 *                              rw_chunk                                     *
 *===========================================================================*/
PRIVATE int rw_chunk(register struct inode *rip, off_t position, int ofs, int nbytes,
		     int rw_flag, char *buff, int seg, int usr)
  /* process one block moving it (maybe partialy) to/from the user space */

/*register struct inode *rip;     pointer to inode for file to be rd/wr */
/*off_t position         	  position in the file */
/*int ofs;               	  offset within a block */
/*int nbytes;                     number of bytes to read or write */
/*int rw_flag;                    READING or WRITING */
/*char *buff;                     virtual address of the user buffer */
/*int seg;                        T or D segment in user space */
/*int usr;                        which user process */
{/* Read or write a part of a block. Used by the read_write function. */

  int result;
  block_t n_block;
  dev_t dev;
  struct buf *block;
  static char zero='\0';

  if ((rip->i_mode & I_TYPE)  ==  I_BLOCK_SPECIAL) {
    n_block = (block_t)(position/BLOCK_SIZE);
    dev = rip->i_zone[0];
  }
  else {
    n_block = read_map(rip,position);
    dev = rip->i_dev;
  }

  if (rw_flag == READING) {
    if (n_block == NO_BLOCK) {
      /* reading from nonexistent block. Return a (part of) block
	full of zeroes */

      /* NIC MOC, NEKDY PREPSAT */
      for (result=0; result<nbytes; result++)
	rw_user(seg,usr,buff+result,1,&zero,TO_USER);
      return nbytes;
    }
    else {
      block = get_block(dev,n_block,NORMAL);
      result = rw_user(seg,usr,buff,nbytes,block->b.b__data+ofs,TO_USER);
      put_block(block,nbytes==BLOCK_SIZE ? FULL_DATA_BLOCK : PARTIAL_DATA_BLOCK);
      if (result < 0)
	return result;
      return nbytes;
    }
  }

  else if (rw_flag == WRITING) {
    if (n_block == NO_BLOCK) /* the block doesn't exist */
     block = new_block(rip,position);
    else
      block = get_block(dev,n_block,nbytes==BLOCK_SIZE ? NO_READ : NORMAL);     /* the block exists */
    result = rw_user(seg,usr,buff,nbytes,block->b.b__data+ofs,FROM_USER);
    put_block(block,nbytes==BLOCK_SIZE ? FULL_DATA_BLOCK : PARTIAL_DATA_BLOCK);
    if (result < 0)
      return result;
    return nbytes;
  }

  return EINVAL;

/* Read or write (part of) a block.
Tudy se cte ze souboru a blokovych zarizeni (znakova primo pres dev_io)

Zjisti, jedna-li se o blokove spec. zarizeni
pokud ano: blok=position/BLOCK_SIZE, dev precti z i_zone[0]

jinak zjisti cislo bloku podle position z i_node.zone[], dev precti z inode
Test, zda blok skutecne existuje (neni soucast diry v soubory):
 pokud ne:
   Read: alokuj blok (get_block()) a napln nulami
   Write: vytvor blok ( new_block() )
 pokud ano:
  Write:over, ma-li byt blok uplne prepsan. Pak get_blok() muze dostat
  flag, ze to ani nema nacitat z disku

 Prenes data z/do uzivatelskeho prostoru (rw_user())
 put_block()*/
}


/*===========================================================================*
 *                              read_map                                     *
 *===========================================================================*/
PUBLIC block_t read_map(register struct inode *rip, off_t position)
/*register struct inode *rip;		ptr to inode to map from */
/*off_t position;                	position in file whose blk wanted */
{/* Given an inode and a position within the corresponding file, locate the
  * block number in which that position is to be found and return it. */

  block_t n_block;
  struct buf *block;

  if (rip->i_size <  position)
    return NO_BLOCK;			  		/* out of file */

  if (position < NR_DZONE_NUM*BLOCK_SIZE)
    return rip->i_zone[(int)position/BLOCK_SIZE];		/* direct block */

  position -= NR_DZONE_NUM*BLOCK_SIZE;
  if (position < (off_t)NR_INDIRECTS*BLOCK_SIZE) {	/* indirect block */
    if (rip->i_zone[NR_DZONE_NUM] == NO_BLOCK)
      return NO_BLOCK;
    block = get_block(rip->i_dev,rip->i_zone[NR_DZONE_NUM],NORMAL);
    n_block = block->b.b__ind[(int)(position/BLOCK_SIZE)];
    put_block(block,INDIRECT_BLOCK);
    return n_block;
  }

  position -= (off_t)NR_INDIRECTS*BLOCK_SIZE;
  if (position < (off_t)NR_INDIRECTS*NR_INDIRECTS*BLOCK_SIZE)
  {						/* double indirect block */
    if (rip->i_zone[NR_DZONE_NUM+1] == NO_BLOCK)
      return NO_BLOCK;
    block = get_block(rip->i_dev,rip->i_zone[NR_DZONE_NUM+1],NORMAL);
    n_block = block->b.b__ind[(int)(position / ((off_t)NR_INDIRECTS*BLOCK_SIZE))];
    put_block(block,INDIRECT_BLOCK);
    if (n_block == NO_BLOCK)
      return NO_BLOCK;
    position %= (off_t)NR_INDIRECTS*BLOCK_SIZE;
    block = get_block(rip->i_dev,n_block,NORMAL);
    n_block = block->b.b__ind[(int)(position/BLOCK_SIZE)];
    put_block(block,INDIRECT_BLOCK);
    return n_block;
  }

  return EINVAL;

/* Given an inode and a position within the corresponding file, locate the
 * block number in which that position is to be found and return it.

7 primych bloku po 1 KB	 					10KB
1 neprimy blok s 256 (1 polozka 4 B) primymi bloky  	256KB
1 neprimy blok druhe urovne s 256 neprimymi bloky  	64MB*/
}


/*===========================================================================*
 *                              write_map                                    *
 *===========================================================================*/
PRIVATE block_t write_map(register struct inode *rip, off_t position)
/*register struct inode *rip;      pointer to inode to be changed */
/*off_t position;                  file address to be mapped */
{/* Allocate a new zone (block), write the change to the zone map of
  * the device, and let the zone belongs to the inode <rip>.
  * If it's neccessary, allocate an indirect or double indirect block too. */

  block_t n_block;
  struct buf *block;

  n_block = read_map(rip, position);
  if (n_block != NO_BLOCK)
    return n_block;

  if (position < NR_DZONE_NUM*BLOCK_SIZE) {		/* direct block */
    n_block = alloc_zone(rip->i_dev);
    rip->i_zone[(int)position/BLOCK_SIZE] = n_block;
    rip->i_dirt = DIRTY;
    return n_block;
  }

  position -= NR_DZONE_NUM*BLOCK_SIZE;
  if (position < (off_t)NR_INDIRECTS*BLOCK_SIZE) {	/* indirect block */
    if (rip->i_zone[NR_DZONE_NUM] == NO_BLOCK) {
      rip->i_zone[NR_DZONE_NUM] = alloc_zone(rip->i_dev);
      rip->i_dirt = DIRTY;
      block = get_block(rip->i_dev,rip->i_zone[NR_DZONE_NUM],INDIRECT_BLOCK);
      clear_block(block);
    }
    else
      block = get_block(rip->i_dev,rip->i_zone[NR_DZONE_NUM],INDIRECT_BLOCK);
    n_block = block->b.b__ind[(int)(position/BLOCK_SIZE)];
    if (n_block == NO_BLOCK) {
      block->b.b__ind[(int)(position/BLOCK_SIZE)] = n_block = alloc_zone(rip->i_dev);
      block->b_dirt = DIRTY;
    }
    put_block(block,INDIRECT_BLOCK);
    return n_block;
  }

  position -= (off_t)NR_INDIRECTS*BLOCK_SIZE;
  if (position < (off_t)NR_INDIRECTS*NR_INDIRECTS*BLOCK_SIZE)
  {						  /* double indirect block */
    if (rip->i_zone[NR_DZONE_NUM+1] == NO_BLOCK) {
      rip->i_zone[NR_DZONE_NUM+1] = alloc_zone(rip->i_dev);
      block = get_block(rip->i_dev,rip->i_zone[NR_DZONE_NUM+1],INDIRECT_BLOCK);
      clear_block(block);
      rip->i_dirt = DIRTY;
    }
    else
      block = get_block(rip->i_dev,rip->i_zone[NR_DZONE_NUM+1],INDIRECT_BLOCK);
    n_block = block->b.b__ind[(int)(position/((off_t)NR_INDIRECTS*BLOCK_SIZE))];
    if (n_block == NO_BLOCK) {
      n_block = alloc_zone(rip->i_dev);
      block->b.b__ind[(int)(position/((off_t)NR_INDIRECTS*BLOCK_SIZE))] = n_block;
      block->b_dirt = DIRTY;
      put_block(block,INDIRECT_BLOCK);
      block = get_block(rip->i_dev,n_block,INDIRECT_BLOCK);
      clear_block(block);
    }
    else {
      put_block(block,INDIRECT_BLOCK);
      block = get_block(rip->i_dev,n_block,INDIRECT_BLOCK);
    }
    position %= (off_t)NR_INDIRECTS*BLOCK_SIZE;
    n_block = block->b.b__ind[(int)(position/BLOCK_SIZE)];
    if (n_block == NO_BLOCK) {
      block->b.b__ind[(int)(position/BLOCK_SIZE)] = n_block = alloc_zone(rip->i_dev);
      block->b_dirt = DIRTY;
    }
    put_block(block,INDIRECT_BLOCK);
    return n_block;
  }

  return NO_BLOCK;
}


/*===========================================================================*
 *                               rw_user                                     *
 *===========================================================================*/
PUBLIC int rw_user(int s, int u, char *vir, int bytes, char *buff,
		   int direction)
/*Prenos dat mezi adr. prostorem (I/D) uzivatele a FS */
/*int s;			D or T space in user proc */
/*int u;			process # */
/*vir_bytes vir;		address within user space */
/*vir_bytes bytes;	# of bytes to transfer */
/*char *buff;		buffer in FS's space */
/*int direction;		TO_USER /FROM_USER */
{/* Call the kernel to read and write user space */

  static message msg;

  msg.m_type = SYS_COPY;
  switch (direction) {
    case TO_USER: msg.SRC_PROC = FS;
		  msg.DST_PROC = u;
		  msg.SRC_SEG = D;
		  msg.DST_SEG = s;
		  msg.SRC_VIR = (long) buff;
		  msg.DST_VIR = (long) vir;
		  msg.BYTE_CNT = bytes;
		  break;
    case FROM_USER: msg.SRC_PROC = u;
		    msg.DST_PROC = FS;
		    msg.SRC_SEG = s;
		    msg.DST_SEG = D;
		    msg.SRC_VIR = (long) vir;
		    msg.DST_VIR = (long) buff;
		    msg.BYTE_CNT = bytes;
		    break;
    default: return EINVAL;
  }
  send_recv(SYSTASK,msg);
  return msg.RET_ERROR_CODE;
}



/*===========================================================================*
 *                              new_block                                    *
 *===========================================================================*/
PUBLIC struct buf *new_block(register struct inode *rip, off_t position)
/* register struct inode *rip;     pointer to inode */
/* off_t position;                 file pointer */
{/* If the position is out of the file or the block have to be overwritten,
  * allocate a new block of the device, and associate it to the inode. */

  block_t n_block;

  n_block = read_map(rip,position);
  if (n_block == NO_ZONE)
    return clear_block(get_block(rip->i_dev,write_map(rip,position),
		       INODE_BLOCK));
  else
    return clear_block(get_block(rip->i_dev,n_block,NO_READ));
}


/*===========================================================================*
 *                             clear_block                                   *
 *===========================================================================*/
PRIVATE struct buf *clear_block(struct buf *block)
{ /* Clears the given block */
  int i;

  for(i=0; i < BLOCK_SIZE; i++)
    block -> b.b__data[i] = 0;
  block -> b_dirt = DIRTY;
  return block;
}
