/*************************************************************************
 * 			super.c - implementation file			 *
 * 		   A part of TUOX file system project '97                *
 *			  by Dalibor Lukas, L 94578                      *
 *  		    who made some corrections in a source		 *
 *		      from Petr Kolonicny .. TUOX FS '96		 *
 *************************************************************************

 * This file manages the super block table and the related data structures,
 * namely, the bit maps that keep track of which zones and which inodes are
 * allocated and which are free.  When a new inode or zone is needed, the
 * appropriate bit map is searched for a free entry.
 *
 * The entry points into this file are
 *   load_bit_maps:   get the bit maps for root or a newly mounted device
 *   unload_bit_maps: write bit maps back to disk after umount
 *   alloc_bit:       somebody wants to allocate a zone or inode; find one
 *   free_bit:        indicate that a zone or inode is available for allocation
 *   get_super:       search the 'superblock' table for a device
 *   mounted:         tells if file inode is on mounted (or ROOT) file system
 *
 *   rw_super:        read a superblock
 */

#include "lib/include/string.h"

#define EXTERN extern

#include "common/com.h"
#include "fs/cache/buf.h"
#include "fs/cache/cache.h"
#include "fs/inode/inode.h"

//#undef EXTERN
//#define EXTERN

#include "fs/inode/super.h"

#undef EXTERN
/* struct super_block super_blocks[NR_SUPERS];	see "super.h" */

/*===========================================================================*
 *                            load_bit_maps                                  *
 *===========================================================================*/
PUBLIC int load_bit_maps(dev_t device)
{ /* Find a superblock belonging to a given device, test cache free buffers,
   * if there is enough space load bitmaps of the device and return TRUE
   * else return FALSE */

  struct super_block *pSuperblock = get_super(device);
  register int i;
  struct buf *pBuffer;

  if (pSuperblock == NIL_SUPER)
    return FALSE;

  /* Try to allocate the neccessary number of buffers for the i-node bitmap */
  for (i = 0; i < pSuperblock->s_imap_blocks; i++)
    if ((pSuperblock->s_imap[i] = get_block(device,2+i,NORMAL)) == NIL_BUF)
      break;

  if (i < pSuperblock->s_imap_blocks) { /* Not enough space */
    for (--i; i > 0; i--) {
      put_block(pSuperblock->s_imap[i],FULL_DATA_BLOCK); /* vracim, nepotrebuju */
      pSuperblock->s_imap[i] = NIL_BUF;
    }
    return FALSE;
  }

  /* Try to allocate the neccessary number of buffers for the zone bitmap */
  for (i = 0; i < pSuperblock->s_zmap_blocks; i++)
    if ((pSuperblock->s_zmap[i] = get_block(device,
		2+pSuperblock->s_imap_blocks+i,NORMAL)) == NIL_BUF)
      break;

  if (i < pSuperblock->s_zmap_blocks) { /* Not enough space */
    for (--i; i > 0; i--) {
      put_block(pSuperblock->s_zmap[i],FULL_DATA_BLOCK); /* vracim, nepotrebuju */
      pSuperblock->s_zmap[i] = NIL_BUF;
    }
    /* v pripade nedostatku cache zrusime i inode bitmapy */
    for (i = 0; i < pSuperblock->s_imap_blocks; i++) {
      put_block(pSuperblock->s_imap[i],FULL_DATA_BLOCK);
      pSuperblock->s_imap[i] = NIL_BUF;
    }
    return FALSE;
  }

  /* Set the first two bits in the inode bitmap as occupated */
  pSuperblock->s_imap[0]->b.b__data[0] |= (unsigned char) 3; /*00000011*/

  /* Set the first bit in the zone bitmap as occupated */
  pSuperblock->s_zmap[0]->b.b__data[0] |= (unsigned char) 1; /*00000001*/
  return TRUE;
}

/*===========================================================================*
 *                           unload_bit_maps                                 *
 *===========================================================================*/
PUBLIC int unload_bit_maps(dev_t device)
{ /* Put back the bufffers saving the i-node and zone bitmaps of a given
   * device. It is used by umount system call. */

  register int i;
  struct super_block *pSuperblock = get_super(device);

  if (pSuperblock == NIL_SUPER)
    return FALSE;

  for (i = 0; i < pSuperblock->s_imap_blocks; i++) {
    put_block(pSuperblock->s_imap[i],MAP_BLOCK);
    pSuperblock->s_imap[i] = NIL_BUF;
  }
  for (i = 0; i < pSuperblock->s_zmap_blocks; i++) {
    put_block(pSuperblock->s_zmap[i],MAP_BLOCK);
    pSuperblock->s_zmap[i] = NIL_BUF;
  }
  return TRUE;
}

/*===========================================================================*
 *                              alloc_bit                                    *
 *===========================================================================*/
PUBLIC bit_t alloc_bit(struct super_block *sp, int map)
{ /* Find the first free bit in a bitmap <map> of a super block <sp>.
   * If there's anyone, mark and return it, else return zero. */

  struct buf **bitmap, *pBlock;
  int blockCount;
  register int i,j,k;
  unsigned char byte;

  /* Set bitmap and block count according to bitmap type */
  if (map == INODE_BITMAP) {
    bitmap = sp->s_imap;
    blockCount = sp->s_imap_blocks;
  }
  else if (map == ZONE_BITMAP) {
    bitmap = sp->s_zmap;
    blockCount = sp->s_zmap_blocks;
  }
  else return 0;	/* invalid <map> argument */

  for (i = 0; i < blockCount; i++) { /* Look for the first free bit */
    pBlock = bitmap[i];
    for (j = 0; j < BLOCK_SIZE; j++)
      if (pBlock->b.b__data[j] != (char)255)
      { /* if the block contains free bits, find the first one in the byte */
	byte = 1;
	for (k = 0; k < 7; k++, byte <<= 1)
	  if ((pBlock->b.b__data[j] & byte) == 0) {
	    pBlock->b.b__data[j] |= byte;
	    pBlock->b_dirt = DIRTY;
	    return i*8*BLOCK_SIZE+8*j+k;
	  }
      }
  }
  return 0; /* no free inode or zone */
}


/*===========================================================================*
 *                              free_bit                                     *
 *===========================================================================*/
PUBLIC void free_bit(struct super_block *sp, int map, bit_t bit_returned)
{ /* Mark a given bit <bit_returned> of a <map> bitmap of a <sp> superblock
   * as free. If the bitmap type is an invalid one, do nothing. */

  int blockNumber = bit_returned / (8*BLOCK_SIZE);
  struct buf *pBuffer;
  int byteNumber;
  char byte;
  register int i;

  if (map == INODE_BITMAP)
    pBuffer = sp->s_imap[blockNumber];
  else if (map == ZONE_BITMAP)
    pBuffer = sp->s_zmap[blockNumber];
  else return;

  bit_returned %= 8*BLOCK_SIZE;       /* bit number relative to the block */
  byteNumber = bit_returned/8; 	      /* byte number relative to the block */
  bit_returned %= 8; 		      /* bit number relative to the byte */
  byte = 1;
  for (i = 7; i > bit_returned; i--)
    byte <<= 1;
  pBuffer->b.b__data[byteNumber] &= (~byte);
  pBuffer->b_dirt = DIRTY;
}

/*===========================================================================*
 *                              get_super                                    *
 *===========================================================================*/
PUBLIC struct super_block *get_super(dev_t device)
{ /* Search the superblock table (super_block[]) for a given <device>.
   * If there's anyone, return it, else return NULL. */

  register int i;
  struct super_block *pSuperblock;

  for (i = 0; i < NR_SUPERS; i++) {
    pSuperblock = super_blocks+i;
    if (pSuperblock->s_dev == device) return pSuperblock;
  }
  return NIL_SUPER;
}


/*===========================================================================*
 *                              mounted                                      *
 *===========================================================================*/
PUBLIC int mounted(register struct inode *rip)
{ /* Return TRUE if a given inode <rip> is in ROOT or another mounted
   * device. */

  return (rip->i_dev == ROOT_DEV  ||  get_super(rip->i_dev) != NIL_SUPER) ?
      TRUE : FALSE;
}


/*===========================================================================*
 *                              rw_super                                   *
 *===========================================================================*/
PUBLIC int rw_super(register struct super_block *sp, int rw_flag)
{ /* If reading is required, try to allocate a cache buffer and copy
   * the superblock from the buffer to <sp>.
   * If writing is asked, copy the superblock <sp> to the allocated buffer. */

  struct buf *pBuffer;

  pBuffer = get_block(sp->s_dev,SUPER_BLOCK,NORMAL);

  if (pBuffer == NIL_BUF)  return FALSE;
  if (rw_flag == READING)
    memcpy(sp,pBuffer,sizeof(d_super_block));
  else if (rw_flag == WRITING) {
    memcpy(pBuffer,sp,sizeof(d_super_block));
    put_block(pBuffer,ZUPER_BLOCK);
  }
  else {
    put_block(pBuffer,ZUPER_BLOCK);
    return FALSE;
  }
  sp->s_dirt = CLEAN;
  return TRUE;
}
