/* super.h - header file
 *
 * Super block table.  The root file system and every mounted file system
 * has an entry here.  The entry holds information about the sizes of the bit
 * maps and inodes.  The s_ninodes field gives the number of inodes available
 * for files and directories, including the root directory.  Inode 0 is
 * on the disk, but not used (the corresponding bit of inode bitmap is always
 * set to "used"
 * The disk layout looks as this:
 *
 *      Item        # blocks
 *    boot block      1
 *    super block     1
 *    inode map     s_imap_blocks
 *    zone map      s_zmap_blocks
 *    inodes        (s_ninodes + 1 + INODES_PER_BLOCK - 1)/INODES_PER_BLOCK
 *   (unused        whatever is needed to fill out the current zone)
 *    data zones    (s_zones - s_firstdatazone) << s_log_zone_size
 *
 * A super_block slot is free if s_dev == NO_DEV.
 */


#ifndef __super_h
#define __super_h

#include "common/const.h"
#include "common/type.h"
#include "fs/common/fsconst.h"
#include "fs/inode/inode.h"

EXTERN struct super_block {
	inode_t s_ninodes;           /* # usable inodes on the minor device */
	zone_t  s_nzones;            /* total device size, including bit maps etc */
	unsigned short s_imap_blocks;/* # of blocks used by inode bit map */
	unsigned short s_zmap_blocks;/* # of blocks used by zone bit map */
	zone_t s_firstdatazone;      /* number of first data zone */
	unsigned short s_log_zone_size;      /* log2 of blocks/zone */
	off_t s_max_size;            /* maximum file size on this device */
	short s_magic;               /* magic number to recognize super-blocks */

	/* The following items are only used when the super_block is in memory. */
	struct buf *s_imap[I_MAP_SLOTS];  /* pointers to in_core inode bit map */
	struct buf *s_zmap[ZMAP_SLOTS];   /* pointers to in_core zone bit map */
	dev_t s_dev;                  /* whose super block is this? */
	struct inode *s_isup;         /* inode for root dir of mounted file sys */
	struct inode *s_imount;       /* (directory) inode mounted on */
	real_time s_time;             /* time of last update */
	int s_rd_only;                /* set to 1 if file sys mounted read only */
	char s_dirt;                  /* CLEAN or DIRTY */
} super_blocks[NR_SUPERS];

typedef struct {
	inode_t s_ninodes;           /* # usable inodes on the minor device */
	zone_t  s_nzones;            /* total device size, including bit maps etc */
	unsigned short s_imap_blocks;/* # of blocks used by inode bit map */
	unsigned short s_zmap_blocks;/* # of blocks used by zone bit map */
	zone_t s_firstdatazone;      /* number of first data zone */
	unsigned short s_log_zone_size;      /* log2 of blocks/zone */
	off_t s_max_size;            /* maximum file size on this device */
	short s_magic;               /* magic number to recognize super-blocks */
} d_super_block;

#define NIL_SUPER (struct super_block *) 0

#define INODE_BITMAP  0
#define ZONE_BITMAP   1

typedef unsigned int bit_t;

PUBLIC int load_bit_maps(dev_t device);
PUBLIC int unload_bit_maps(dev_t device);
PUBLIC bit_t alloc_bit(struct super_block *sp, int map);
PUBLIC void free_bit(struct super_block *sp, int map, bit_t bit_returned);
PUBLIC struct super_block *get_super(dev_t device);
PUBLIC int mounted(register struct inode *rip);
PUBLIC int rw_super(register struct super_block *sp, int rw_flag);

#endif

