/*************************************************************************
 *                          TUOX - knihovny                              *
 *************************************************************************
 * Soubor:    dir.c                                                      *
 * Popis:     Obsahuje tyto funkce:                                      *
 *************************************************************************            				                                 *
 *    ACCESS     testuje prava souboru                                   *
 *    CHDIR      zmena pracovniho adresare                               * 
 *    CHMOD      zmena prav souboru                                      *
 *    CHOWN      zmena vlastnika souboru                                 *
 *    CHROOT     zmena korenoveho adresare                               *
 *    CLOSE      zavreni souboru                                         *
 *    CREAT      vytvori soubor                                          * 
 *    FSTAT      vyplnuje struct stat podle file descriptoru             *
 *    LINK       vytvori link                                            *
 *    LSEEK      zmeni aktualni pozici v souboru na stanovene misto      *
 *               vzhledem k pocatku, aktualni pozici nebo konci souboru  *
 *    MKNOD      vytvori specialni soubor                                *
 *    OPEN       otevre soubor                                           *
 *    READ       cte ze souboru                                          *
 *    STAT       vyplnuje struct stat podle jmena souboru                *
 *    STIME      nastavi cas                                             *
 *    TIME       zjisti aktualni cas                                     *
 *    TIMS       zjisti casy procesu                                     * 
 *    UMASK      zmena defaultni masky prav                              *
 *    UNLINK     zrusi link                                              *
 *    UTIME      meni cas v souboru                                      *
 *    WRITE      zapisuje do souboru                                     *
 *                                                                       *
 *    MOUNT      provede namontovani noveho file systemu do adres. stromu*
 *    UMOUNT     provede odmontovani file systemu z adresaroveho stromu  *
 *                                                                       *
 *    SYNC        provede flush cache pred shutdown                      *
 *************************************************************************
 * Autor:     Martin Vojvodik	                                         *
 *	      Bosak Rostislav                                            *
 *            Hurdes Jaroslav                                            *
 *            Perina Martin                                              *
 * 							                 *
 * Preprac:   Daniel Tvrdy 			                         *
 *            Ivo Orlik	                                        (c) 1997 *
 *************************************************************************/

#include "LIB\include\dir.h"
#include "COMMON\callnr.h"
#include "COMMON\const.h"
#include "COMMON\stat.h"       /* ??? pro prikaz chmod - maska */
#include "kernel\makra.h"
#include "FS\common\fsmes.h"         /* hlavickovy soubor FS */


/***********************************************************************/

int access (const char *_path, int _mode)
{
  message m;

  m.m_type = C_ACCESS;              /* FS - testuje prava souboru */
  m.fs_name = (char *) _path;
  m.fs_mode = _mode;                /* testovana maska prav souboru */
  send_recv(FS,m);
                                            
  return m.fs_errcode;    /* vracime pripadny chybovy kod */       
}


/***********************************************************************/

int chdir (const char *_path)
{
  message m;

  m.m_type = C_CHDIR;               /* FS - zmena pracovniho adresare */
  m.fs_name = (char *) _path;       /* ukazatel na jmeno noveho prac. adresare */
  send_recv(FS,m);
                                            
  return m.fs_errcode;   /* vracime pripadny chubovy kod */
}


/***********************************************************************/

int chmod (const char *_path, int _mode)
{
  message m;

  m.m_type = C_CHMOD;             /* FS - zmena prav souboru */
  m.fs_name = (char *) _path;
  m.fs_mode = _mode;              /* set maska */
  send_recv(FS,m);

  return m.fs_errcode;    /* vracime pripadny chybovy kod */
}


/***********************************************************************/

int chown (const char *_path, int _uid, int _gid)
{
  message m;

  m.m_type = C_CHOWN;            /* FS - zmena vlastnika souboru */
  m.fs_name = (char *) _path;  /* ukazatel na jmeno souboru */
  m.fs_owner = _uid;           /* identifikacni cislo uzivatele - uid */    
  m.fs_group = _gid;           /* identifikacni cislo skupiny - gid */
  send_recv(FS,m);

  return m.fs_errcode;      /* vracime pripadny chybovy kod */
}


/***********************************************************************/

int chroot (const char *_newroot)
{
  message m;

  m.m_type = C_CHROOT;              /* FS - zmena korenoveho adresare */
  m.fs_name = (char *) _newroot;  /* ukazatel na jmeno adresare */
  send_recv(FS,m);

  return m.fs_errcode;    /* vracime pripadny chybovy kod */
}


/***********************************************************************/

int close (int _handle)
{
  message m;

  m.m_type = C_CLOSE;           /* FS - zavre soubor */
  m.fs_fd = _handle;            /* file descriptor */
  send_recv(FS,m);

  return m.fs_errcode;   /* vracime pripadny chybovy kod */
}


/***********************************************************************/

int creat (const char *_path, int _mode)
{
  message m;

  m.m_type = C_CREAT;             /* FS - vytvori novy soubor */
  m.fs_name = (char *) _path;     /* ukazatel na jmeno souboru */
  m.fs_mode = _mode;              /* prava souboru */
  send_recv(FS,m);

  return m.fs_errcode;   /* vracime file discriptor nebo chybovy kod */
}


/***********************************************************************/

int     fstat (int _handle, struct stat *_buffer)
{
  message m;

  m.m_type = C_FSTAT;                /* FS - vypnuje struct stat podle file descriptoru */
  m.fs_fd = _handle;                 /* file descriptor souboru */
  m.fs_buf = (char *) _buffer;       /* pointer na vystupni buffer */
  send_recv(FS,m);

  if (m.fs_errcode == 0)
  {
/*    _buffer = m.fs_buf;   /* vyplnena struktura stat */
    return 0;               /* vse v poradku */
  }
  return m.fs_errcode;    /* nebo chybovy kod */
}


/***********************************************************************/

int link (const char *_path, const char *_link)
{
  message m;

  m.m_type = C_LINK;	         /* FS - vytvori link */
  m.fs_name1 = (char *)_path;    /* ukazatel na jmeno souboru */
  m.fs_name2 = (char *)_link;    /* ukazatel na jmeno linku */
  send_recv(FS,m);

  return m.fs_errcode;   /* vracime pripadny chybovy kod */
}


/***********************************************************************/

int lseek (int _handle, int _offset, int _whence)
{
  message m;

  m.m_type = C_LSEEK;             /* FS - zmeni aktualni pozici v soub. */
  m.fs_fd = _handle;              /* file descriptor */
  m.fs_offset = _offset;          /* offset */
  m.fs_whence = _whence;          /* vztazny bod pro posun */
  send_recv(FS,m);
                                             
  if (m.fs_errcode == 0)
    return m.fs_pos;     /* vracime novou pozici v souboru */ 
  return m.fs_errcode;   /* chybovy kod */
}


/***********************************************************************/

int mknod (const char *_specf, int _mode, int _addr)
{
  message m;

  m.m_type = C_MKNOD;              /* FS - vytvori specialni soubor */
  m.fs_name = (char *) _specf;     /* ukazatel na jmeno specialniho souboru */
  m.fs_mode = _mode;               /* prava souboru */
  m.fs_major_minor = _addr;	   /* major a minor device number */
  send_recv(FS,m);

  return m.fs_errcode;      /* vracime pripadny chybovy kod */
}


/***********************************************************************/

int open (const char *_path, int _mode)     /* ??? mode (rw_flag) */
{
  message m;

  m.m_type = C_OPEN;              /* FS - otevre soubor */
  m.fs_name = (char *) _path;     /* ukazatel na jmeno souboru */
  m.fs_mode = _mode;              /* read / write flag */
  send_recv(FS,m);
  return m.fs_fd;               /* vracime file discriptor nebo err code */
}


/***********************************************************************/

int read (int _handle, void *_buf, unsigned _len)
{
  message m;

  m.m_type = C_READ;               /* FS - cte ze souboru */
  m.fs_fd = _handle;               /* file descriptor */
  m.fs_buffer = (char *) _buf;     /* buffer */
  m.fs_nbytes = _len;              /* pocet bytu, ktere se maji precist */
  m.fs_seg = D;
  send_recv(FS,m);

  return m.fs_errcode;      /* vratit pocet prectenych bajtu nebo errcode */
}


/***********************************************************************/

int stat (const char *_path, struct stat *_buffer)
{
  message m;

  m.m_type = C_STAT;                /* FS - vyplnuje struct stat podle jmena souboru */
  m.fs_name = (char *) _path;       /* pointer na nazev adresare */
  m.fs_buff = (char *) _buffer;     /* pointer na vyst buffer */
  send_recv(FS,m);

  if (m.fs_errcode == 0)
  {
/*    _buffer = m.fs_buff;     /* vratit vyplnenou strukturu struct stat */
    return 0;                  /* vracime priznak OK */
  }
  return m.fs_errcode;         /* nebo chybovy kod */
}


/***********************************************************************/

int stime (char *_tp)
{
  message m;

  m.m_type = C_STIME;                 /* FS - nastavi cas */
  m.fs_tp = (char *) _tp;
  send_recv(FS,m);

  return m.fs_errcode;    /* !!! return rw_user */
}


/***********************************************************************/

long time (long *_tp)
{
  message m;

  m.m_type = C_TIME;                 /* FS - zjisti aktualni cas */
  m.fs_tp = (char *) _tp;
  send_recv(FS,m);
                                            
  if (m.fs_errcode == 0)
  {
    return 0;    /* vracime aktualni cas */
  }
  return (long) m.fs_errcode;    /* nebo chybovy kod */
}


/***********************************************************************/

int tims (long *_buf)
{
  message m;

  m.m_type = C_TIMES;                   /* FS - zjisti casy procesu */
  m.fs_buf = _buf;
  send_recv(FS,m);
                                            
  return m.fs_errcode;    /* return rw_user */
}


/***********************************************************************/

unsigned int umask (unsigned int _mode)
{
  message m;

  m.m_type = C_UMASK;            /* FS - zmena defaultni masky prav */
  m.fs_mode = _mode;             /* nova maska */
  send_recv(FS,m);

  return m.fs_oldmode;      /* vraci starou masku */
}


/***********************************************************************/

int unlink (const char *_link)
{
  message m;

  m.m_type = C_UNLINK;                /* FS - zrusi link */
  m.fs_name = (char *) _link;	      /* ukazatel na jmeno souboru */
  send_recv(FS,m);

  return m.fs_errcode;    /* vracime pripadny chybovy kod */
}


/***********************************************************************/
/*
int utime (const char *_path, struct utimbuf timep)
{
  message m;

  m.m_type = C_UTIME;                  // FS - zmena casu souboru
  m.fs_namet = (char *) _path;
  m.fs_timep = timep; // !!!!!
  send_recv(FS,m);

  return m.fs_errcode;     // vracime chybovy kod
}
*/

/***********************************************************************/

int write (int _handle, void *_buf, unsigned _len)
{
  message m;

  m.m_type = C_WRITE;              /* FS - zapisuje do souboru */
  m.fs_fd = _handle;               /* file descriptor */
  m.fs_buffer = (char *) _buf;     /* buffer */
  m.fs_nbytes = _len;              /* pocet bytu, ktere se maji zapsat */
  m.fs_seg = D;                 /* typ segmentu uzivatelske oblasti */
  send_recv(FS,m);

  return m.fs_errcode;      /* vratit pocet zapsanych bajtu nebo errcode */
}



/***********************************************************************/
/***********************************************************************/


int mount (const char *_specf, const char *_path, int _readonly)
{
  message m;

  m.m_type = C_MOUNT;          
  m.m1_p1 = (char *) _specf;   
  m.m1_p2 = (char *) _path;    
  m.m1_i1 = _readonly;         
  send_recv(FS,m);

  return m.fs_errcode;   
}


/***********************************************************************/

int umount (const char *_specf)
{
  message m;

  m.m_type = C_UMOUNT;
  m.m1_i1 = (char *) _specf;  
  send_recv(FS,m);

  return m.fs_errcode;   
}




/***********************************************************************/

void sync (void)
{
  message m;

  m.m_type = C_SYNC;                  /* FS - flush cache pred shutdown */
  send_recv(FS,m);
                                            
  /* nevracime nic */
}

