/*************************************************************************
 *                          TUOX - knihovny                              *
 *************************************************************************
 * Soubor:    process.c                                                  *
 * Popis:     Obsahuje tyto funkce pro manipulaci s procesy:             *
 *************************************************************************
 *    EXIT      ukonci volajici proces, navratova hodnota = status      *
 *    BRK        posune hranici gapu                                     *
 *    WAIT       pocka na ukonceni vsech detskych (child) procesu        *
 *    WAITPID    pocka na ukonceni procesu se zadanym pid                *
 *    ALARM      nastavi spusteni alarmu na cas, ktery nastane po uply-  *
 *               nuti zadaneho poctu sekund                              *
 *    PAUSE      ?? nevim ??                                             *
 *    FORK       rozdeli proces na dva totozne, startujici prikazem      *
 *               nasledujicim za fork()                                  * 
 *    KILL       ukonci proces se zadanym pid poslanim signalu           *
 *    GETPID     vrati pid volajiciho procesu                            *
 *    GETUID     ulozi do __ruid a __euid relativni a efektivni UID      *
 *               procesu __pid (__pid == 0 svuj)                         *
 *    GETGID     ulozi do __rgid a __egid relativni a efektivni GID      *
 *               procesu __pid (__pid == 0 svuj)                         *
 *    SETUID     nastavi efektivni UID volajiciho procesu na __uid,      *
 *               vyzaduje prava SUPER_USER                               *
 *    SETGID     nastavi efektivni GID volajiciho procesu na __gid,      *
 *               vyzaduje prava SUPER_USER                               *
 *    GETPGRP    vrati GROUP procesu __pid                               *
 *    SETPGRP    nastavi GROUP procesu __pid na __group, vyzaduje prava  *
 *               SUPER_USER                                              * 
 *    GETMPROC   zkopiruje tabulku procesum_proc do mista daneho ukaza-  *
 *               telem __where                                           * 
 *    SIGNAL     nastavuje obsluznou funkci signalu                      * 
 *************************************************************************
 * Autor:     Martin Vojvodik	                                         *
 *	      Bosak Rostislav                                            *
 *            Hurdes Jaroslav                                            *
 *            Perina Martin                                              *
 * 							                 *
 * Preprac:   Daniel Tvrdy 			                         *
 *            Ivo Orlik	                                        (c) 1997 *
 *************************************************************************/



#include "COMMON\callnr.h"
#include "COMMON\type.h"
#include "COMMON\const.h"
#include "COMMON\errno.h"
#include "COMMON\signal.h"      /* pro nastaveni signalu */
#include "COMMON\stat.h"        /* pro amode (access, chmod ...) */
#include "MM\mm.h"              /* hlavickovy soubor MM */
#include "kernel\makra.h"
#include "LIB\include\process.h"

static message m;

/**************************************************************************
 * Funkce: exit                                                           *
 **************************************************************************
 * Popis: ukonci volajici proces, navratova hodnota = status              *
 * Funkce vraci: nic                                                      *
 **************************************************************************/

void exit(int status)
{
  m.m_type = C_EXIT;
  m.MM_EXITSTATUS = status;
  send_recv(MM,m);
}



/**************************************************************************
 * Funkce: brk                                                            *
 ************************************************************************** 
 * Popis: posune hranic gapu                                              *
 * Funkce vraci: - NULL - pokud nastala chyba                             *
 *               - jinak vracime adresu, kde je konec alokovatelne casti  *
 **************************************************************************/

void *brk(int __incr)
{
  m.m_type = C_BRK;
  m.MM_BRKINCR = __incr;  /* relativni adresa o kolik posunout hranici gapu (nahoru nebo dolu) */
  send_recv(MM,m);

  if (m.MM_ERRORNO != 0)
    return NULL;          /* pokud nastala chyba vracime NULL */

  return (void *) m.MM_BRKADDR;  /* vracime adresu, kde je konec alokovatelne casti   */
				 /* od zacatku zjisteneho prvnim volanim brk funkce */
}



/**************************************************************************
 * Funkce: wait                                                           *
 ************************************************************************** 
 * Popis: pocka na ukonceni vsech child procesu                           *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - jinak vracime pid procesu, ktery skoncil               *
 **************************************************************************/

int wait (void)
{
  m.m_type = C_WAITPID;
  m.MM_WAITFOR = -1;
  m.MM_WAITOPT = WNOHANG;
  send_recv(MM,m);

  if (m.MM_ERRORNO != 0)
    return m.MM_ERRORNO;    /* vracime chybovy kod */
  return m.MM_EXITEDPID;      /* vracime pid procesu, ktery skoncil */
}



/**************************************************************************
 * Funkce: waitpid                                                        *
 **************************************************************************
 * Popis: pocka na ukonceni procesu se zadanym pid                        *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - jinak vracime pid procesu, ktery skoncil               *
 **************************************************************************/

int  waitpid (pid_t __pid, int __options)
{
  m.m_type = C_WAITPID;
  m.MM_WAITFOR = __pid;       /* pid na koho cekat */
  m.MM_WAITOPT = __options;   /* pripadne dalsi nastaveni */
  send_recv(MM,m);

  if (m.MM_ERRORNO != 0)
    return m.MM_ERRORNO;      /* vracime chybovy kod */ 
  return m.MM_EXITEDPID;        /* ??? mozno vracet jeste exitstatus, a terminate status */
}


/**************************************************************************
 * Funkce: alarm                                                          *
 ************************************************************************** 
 * Popis: nastavi spusteni alarmu na cas, ktery nastane po uplynuti       *
 *        zadaneho poctu sekund                                           *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - 0 - vse v poradku                                      *
 **************************************************************************/

int  alarm (int __seconds)
{
  m.m_type = C_ALARM;
  m.MM_ALARMSECOND = __seconds;   /* pocet sekund do prichodu alarmu */
  send_recv(MM,m);

  return m.MM_ERRORNO;            /* vracime pripadny chybovy kod */
}


/**************************************************************************
 * Funkce: pause                                                          *
 ************************************************************************** 
 * Popis:                                                                 *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - 0 - vse v poradku                                      *
 **************************************************************************/

int  pause()
{
  m.m_type = C_PAUSE;
  send_recv(MM,m);

  return m.MM_ERRORNO;      /* vracime pripadny chybovy kod */
}



/**************************************************************************
 * Funkce: fork                                                           *
 ************************************************************************** 
 * Popis: rozdeli proces na dva totozne, startujici prikazem nasledujicim *
 *        za fork()                                                       *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - jinak pid noveho detskeho procesu                      *
 **************************************************************************/

pid_t  fork()
{
  pid_t my_pid, child_pid;

  m.m_type = C_FORK;
  send_recv(MM,m);
  child_pid = m.MM_FORKEDCHILD;    /* pid noveho detskeho procesu */
  my_pid = getpid();               /* muj vlastni pid */

  if (child_pid == my_pid)         /* jestli nedoslo ke zmene vrat nulu */
    return 0;

  if (m.MM_ERRORNO != 0)
    return m.MM_ERRORNO;           /* vracime chybovy kod */
  return child_pid;                /* jinak vrat pid noveho detskeho procesu */
}



/**************************************************************************
 * Funkce: kill                                                           *
 **************************************************************************
 * Popis: ukonci proces se zadanym pid poslanim signalu                   *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - 0 - vse v poradku                                      *
 **************************************************************************/

int     kill(pid_t __pid, int __sig)
{
  m.m_type = C_KILL;
  m.MM_PID = __pid;	     /* pid procesu, ktery chceme ukoncit */
  m.MM_SIGNAL = __sig;       /* signal, ktery mu chceme poslat */
  send_recv(MM,m);

  return m.MM_ERRORNO;       /* vracime pripadny chybovy kod */
}


/**************************************************************************
 * Funkce: getpid                                                         *
 ************************************************************************** 
 * Popis: vrati pid volajiciho procesu                                    *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - jinak vracime pid volajiciho procesu                   *
 **************************************************************************/

pid_t  getpid(void)
{
  m.m_type = C_GETPID;
  send_recv(MM,m);

  if (m.MM_ERRORNO != 0)
    return m.MM_ERRORNO;     /* vracime chybovy kod */
  return m.MM_GPID;          /* vracime pid volajiciho procesu */
}


/**************************************************************************
 * Funkce: getuid                                                         *
 ************************************************************************** 
 * Popis: ulozi do __ruid a __euid skutecne a efektivni UID procesu __pid *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - jinak vracime realne uid                               *
 **************************************************************************/

uid_t   getuid(pid_t __pid, uid_t *__ruid, uid_t *__euid)
{
  m.m_type = C_GETUID;
  m.MM_GETIDPID = __pid;      /* pid procesu na ktery se dotazujeme (==0 svuj) */
  send_recv(MM,m);
  *__ruid = m.MM_GRUID;        /* realne UID */
  *__euid = m.MM_GEUID;        /* efektivni UID */

  if (m.MM_ERRORNO != 0)
    return m.MM_ERRORNO;      /* vracime chybovy kod */
  return *__ruid;             /* vracime realne UID */
}



/**************************************************************************
 * Funkce: getgid                                                         *
 ************************************************************************** 
 * Popis: ulozi do __rgid a __egid skutecne a efektivni GID procesu __pid *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - jinak vracime realne gid                               *
 **************************************************************************/

gid_t  getgid(pid_t __pid, gid_t *__rgid, gid_t *__egid)
{
  m.m_type = C_GETGID;
  m.MM_GETIDPID = __pid;     /* pid procesu na ktery se dotazujeme (==0 svuj) */
  send_recv(MM,m);
  *__rgid = m.MM_GRGID;       /* realne GID */
  *__egid = m.MM_GEGID;       /* efektivni GID */

  if (m.MM_ERRORNO != 0)
    return m.MM_ERRORNO;     /* vracime chybovy kod */
  return *__rgid;            /* vracime realny GID */
}


/**************************************************************************
 * Funkce: setuid                                                         *
 ************************************************************************** 
 * Popis: nastavi efektivni UID volajiciho procesu na __uid,              *
 *        vyzaduje prava SUPER_USER                                       *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - 0 - vse v poradku                                      *
 **************************************************************************/

int  setuid (uid_t __uid)
{
  m.m_type = C_SETUID;
  m.MM_SUID = __uid;	        /* nova hodnota UID */
  send_recv(MM,m);

  return m.MM_ERRORNO;          /* vracime pripadny chybovy kod */
}


/**************************************************************************
 * Funkce: setgid                                                         *
 ************************************************************************** 
 * Popis: nastavi efektivni GID volajiciho procesu na __gid,              *
 *        vyzaduje prava SUPER_USER                                       *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - 0 - vse v poradku                                      *
 **************************************************************************/

int  setgid (gid_t __gid)
{
  m.m_type = C_SETGID;
  m.MM_SGID = __gid;        /* nova hodnota GID, kterou chceme nastavit */
  send_recv(MM,m);

  return m.MM_ERRORNO;      /* vracime pripadny chybovy kod */
}


/**************************************************************************
 * Funkce: getpgrp                                                         *
 ************************************************************************** 
 * Popis: vrati GROUP procesu __pid                                       *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - jinak vracime group procesu                            *
 **************************************************************************/

int  getpgrp (pid_t __pid)
{
  m.m_type = C_GETPGRP;
  m.MM_GPGRPPID = __pid;      /* pid procesu, u ktereho nas zajima GROUP (==0 zadajici proces) */
  send_recv(MM,m);

  if (m.MM_ERRORNO != 0)
    return m.MM_ERRORNO;     /* vracime pripadny chybovy kod */
  return m.MM_PGRP;          /* vracime GROUP procesu */
}


/**************************************************************************
 * Funkce: setpgrp                                                         *
 ************************************************************************** 
 * Popis: nastavi GROUP procesu __pid na __group,vyzaduje prava SUPER_USER*
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - 0 - vse v poradku                                      *
 **************************************************************************/

int  setpgrp (pid_t __pid, int __group)
{
  m.m_type = C_SETPGRP;
  m.MM_SPGRPPID = __pid;	/* proces, kteremu budeme menit GROUP (==0 zadajici proces) */
  m.MM_SPGRPVALUE = __group;    /* nova hodnota GROUP */
  send_recv(MM,m);

  return m.MM_ERRORNO;	        /* pripadny chybovy kod */
}


/**************************************************************************
 * Funkce: getmproc                                                       *
 ************************************************************************** 
 * Popis: zkopiruje tabulku procesu m_proc do mista daneho ukazatelem     *
 *        __where                                                         *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - jinak vracime velikost tabulky mproc                   *
 **************************************************************************/
/* opt = __MPROCCOPY
 *  zkopiruje tabulku procesu m_proc do mista daneho ukazatelem __where
 *  (misto musi byt v datovem segmentu volajiciho procesu (staticky
 *  alokovano)).
 * opt = __MPROCSIZE
 *  vrati pouze velikost, nekopiruje
 * Vraci velikost tabulky MPROC */

int  getmproc (void *__where, int __opt)
{
  m.m_type = C_GETPROCTBL;
  m.MM_GETMPROCPTR = __where;     /* ukazatel kam se ma kopirovat */
  m.MM_GETMPROCOPT = __opt;
  send_recv(MM,m);

  if (m.MM_ERRORNO != 0)
    return m.MM_ERRORNO;         /* pripadny chybovy kod */
  return m.MM_GETMPROCSIZE;      /* velikost tabulky mproc */
}

/**************************************************************************
 * Funkce: signal                                                         *
 ************************************************************************** 
 * Popis: nastavi signal, pokud func == NULL, pak nastavi signal IGNORE   *
 * Funkce vraci: - chybovy kod - pokud nastala chyba                      *
 *               - 0 - vse v poradku                                      *
 **************************************************************************/

void cdecl (* cdecl signal(int __sig, void cdecl (* func)())) (int)
{
  m.MM_SIGNAL = __sig;            /* cislo signalu */
  if(func != NULL)
    m.MM_PTRFUNC = func;          /* ukazatel na uzivatelskou funkci */
  else
    m.MM_PTRFUNC = SIG_IGN;       /* nastavit signal IGNORE */

  send_recv(MM,m);

  return m.MM_ERRORNO;              /* vracime pripadny chybovy kod */
}




/* ???? takova funkce neexistuje */
/* shozeni celeho systemu
int     shutdown()
{
  message m_u;

  m_u.m_type = SHUTDOWN;
  send_recv(MM,m_u);

  return MM_ERRORNO;
}                         */
