/*
 ****************************************************************************
 * soubor: alloc.c                                                          *
 * autor:  Marcel Kruzel, Zdenek Sawa, Milan Gvozdik                        *
 * ladici: Rojicek Radim, Sela Jaroslav                                     *
 ****************************************************************************
*/

/*
   Tato cast tuoxu je dodelana.
   Byla dokonce otestovana na rozumnem priklade.
   Zde hledejte chyby az jako posledni instance.


   This file is concerned with allocating and freeing arbitrary-size blocks of
 * physical memory on behalf of the FORK and EXEC system calls.  The key data
 * structure used is the hole table, which maintains a list of holes in memory.
 * It is kept sorted in order of increasing memory address. The addresses
 * it contains refer to physical memory, starting at absolute address 0
 * (i.e., they are not relative to the start of MM).  During system
 * initialization, that part of memory containing the interrupt vectors,
 * kernel, and MM are "allocated" to mark them as not available and to
 * remove them from the hole list.
 *
 * The entry points into this file are:
 *   alloc_mem: allocate a given sized chunk of memory
 *   free_mem:  release a previously allocated chunk of memory
 *   mem_init:  initialize the tables when MM start up
 *   max_hole:  returns the largest hole currently available
 */


/* Trideni dle velikosti bloku padlo protoze se ukazalo, ze to dela
   problemy pri uvolneni bloku atd.
   Pokud chceme allokovat vzdy nejmensi kousek musime ho vyhledat v alloc_mem.
*/

#include "mm\mm.h"
#include "mm\mm_const.h"
#include "mm\alloc.h"

#define NIL_HOLE (struct hole *) 0

PRIVATE struct hole {
  phys_clicks h_base;           /* where does the hole begin? */
  phys_clicks h_len;            /* how big is the hole? */
  struct hole *h_next;          /* pointer to next entry on the list */
} hole[NR_HOLES];

PRIVATE struct hole *hole_head,*free_slots;    

/*===========================================================================*
 *                              between                                      *
 *===========================================================================*
Prida diru pred behined (ne tesne). Pouzito ve free_mem.
*/
PRIVATE struct hole *between(struct hole *behined,phys_clicks base,phys_clicks clicks)
{
  struct hole *aHole=free_slots;
  free_slots=free_slots->h_next;
  aHole->h_next=behined;
  aHole->h_base=base;
  aHole->h_len=clicks;
  return aHole;
};

/*===========================================================================*
 *                              alloc_mem                                    *
 *===========================================================================*/
PUBLIC phys_clicks alloc_mem(clics)
phys_clicks clics;             /* amount of memory requested */
{
/* Allocate a block of memory from the free list. The block
 * consists of a sequence of contiguous bytes, whose length in paragraphs is
 * given by 'clicks'.  A pointer to the block is returned.  The block is
 * always on a click boundary.  This procedure is called when memory is
 * needed for FORK or EXEC.
 */
  struct hole *aHole=hole_head,*aPrevHole=NIL_HOLE;
  phys_clicks holeAddr=0;
  while (aHole!=NIL_HOLE)
  {
    if (aHole->h_len>=clics)
    {
      holeAddr=aHole->h_base;      /* Prideleni pameti */
      aHole->h_len-=clics;
      aHole->h_base+=clics;
      if (aHole->h_len==0)         /* Cela dira zaplnena */
      {
	if (aPrevHole==NIL_HOLE)   /* Allokujeme celou diru na zacatku seznamu */
	  hole_head=aHole->h_next;
	else                       /* Allokujeme celou diru nekde uvnitr seznamu */
	  aPrevHole->h_next=aHole->h_next;
        aHole->h_next=free_slots;
	free_slots=aHole;
      } else {};                   /* Zde eventuelni trideni */
      aHole=NIL_HOLE;              /* Ukonceni cyklu */
    } else
    {
      aPrevHole=aHole;             /* Posunuti v seznamu */
      aHole=aHole->h_next;
    };
  };
  return holeAddr;
};

/*===========================================================================*
 *                              free_mem                                     *
 *===========================================================================*/
PUBLIC void free_mem(base, clicks)
vir_clicks base;               /* base address of block to free */
phys_clicks clicks;             /* number of clicks to free */
{
/* Return a block of free memory to the hole list.  The parameters tell where
 * the block starts in physical memory and how big it is.  The block is added
 * to the hole list.  If it is contiguous with an existing hole on either end,
 * it is merged with the hole or holes.
 */

  struct hole *aHole=hole_head,*aPrevHole=NIL_HOLE;
  if (free_slots==NIL_HOLE){}/* panic("MMError:Nedostatek volneho mista pro sloty volnych der!");*/else
  if (aHole==NIL_HOLE)   /* Zadna dira v systemu */
  {
    hole_head=free_slots;
    free_slots=free_slots->h_next;
    hole_head->h_base=base;
    hole_head->h_len=clicks;
    hole_head->h_next=NIL_HOLE;    
  }else
  { 
    while (aHole!=NIL_HOLE)     /* Hledame, kde to vrazit mezi diry dle adres. */
    { 
      if (aHole->h_base>base) break;
      aPrevHole=aHole;
      aHole=aHole->h_next;
    };  /* Vysledkem je: aPrevHole ... za nim je uvolnovany blok
                         aHole ....... pred nim je uvolnovany blok */
        /* Test zda mezi dvemi derami */
    if (aPrevHole==NIL_HOLE)    /* Pred prvni dirou */
    {
      if (hole_head->h_base-clicks==base)  /* Tesne pred prvni dirou */
      {
        hole_head->h_base-=clicks;
        hole_head->h_len+=clicks;
      }else                               /* Netesne pred prvni dirou */
      {
        hole_head=between(hole_head,base,clicks);
      };     
    }else  
    {
      if (aHole==NIL_HOLE)      /* Za posledni diru */
      {
	if (aPrevHole->h_base+aPrevHole->h_len==base) /* Tesne za posledni diru */
        {
	  aPrevHole->h_len+=clicks;
        }else                                        /* Netesne za posledni diru */
        {
          aPrevHole->h_next=between(NIL_HOLE,base,clicks);
        };
      }else                    /* Mezi dvemi dirami */
      { 
        if (aPrevHole->h_base+aPrevHole->h_len==base) /* Tesne za prvni dirou */
        {
          if (aHole->h_base-clicks==base)             /* Natesno mezi prvni a druhou */
          {
	    aPrevHole->h_len+=aHole->h_len+clicks;
            aPrevHole->h_next=aHole->h_next;
	    aHole->h_next=free_slots;
            free_slots=aHole; 
	  }else                                      /* Pouze tesne za prvni */
          {
            aPrevHole->h_len+=clicks;
          };
        }else
        {
          if (aHole->h_base-clicks==base)             /* Pouze tesne pred druhou dirou */
          {
            aHole->h_base-=clicks;
	    aHole->h_len+=clicks;
          }else                 /* Ani pred druhou ani za prvni dirou - mezi dvema */
	  {
            aPrevHole->h_next=between(aPrevHole->h_next,base,clicks);
          };
        };
      };
    };
  };
};

/*===========================================================================*
 *                              max_hole                                     *
 *===========================================================================*/
PUBLIC phys_clicks max_hole()
{
/* Scan the hole list and return the largest hole. */
/* Hleda nejvetsi diru */
  phys_clicks max_hole=0;                     /* velikost maximalni diry */
  struct hole *aHole=hole_head;               /* prohledavani seznamu    */
  while (aHole!=NIL_HOLE) 
  {
    if (aHole->h_len>max_hole) max_hole=aHole->h_len;
    aHole=aHole->h_next;
  };
  return max_hole;
}


/*===========================================================================*
 *                              mem_init                                     *
 *===========================================================================*/
PUBLIC void mem_init(phys_clicks base, phys_clicks size)
 /* base je prvni volny segment, size je jeho delka v paragrafech (clicks)*/
{
/* Initialize hole lists.  There are two lists: 'hole_head' points to a linked
 * list of all the holes (unused memory) in the system; 'free_slots' points to
 * a linked list of table entries that are not in use.  Initially, the former
 * list has one entry for each chunk of physical memory, and the second
 * list links together the remaining table slots.  As memory becomes more
 * fragmented in the course of time (i.e., the initial big holes break up into
 * smaller holes), new table slots are needed to represent them.  These slots
 * are taken from the list headed by 'free_slots'.
 */
  unsigned int i;
  hole_head=hole;                      /* Vytvoreni seznamu der */
  hole_head->h_base=base;              /* Na zacatku jedina velka dira */
  hole_head->h_len=size;
  hole_head->h_next=NIL_HOLE;
				       /* Vytvoreni seznamu volnych slotu */
  for (i=1;i<NR_HOLES-1;i++) hole[i].h_next=hole+i+1;
  hole[NR_HOLES-1].h_next=NIL_HOLE;    /* Konec seznamu volnych slotu */
  free_slots=hole+1;                   /* Seznam zacina druhym prvkem pole */
}
