/*
 ****************************************************************************
 * soubor: exec.c                                                           *
 * autor:  Marcel Kruzel, Zdenek Sawa, Milan Gvozdik                        *
 * ladici: Rojicek Radim, Sela Jaroslav                                     *
 ****************************************************************************
*/

/* This file handles the EXEC system call.  It performs the work as follows:
 *    - see if the permissions allow the file to be executed
 *    - read the header and extract the sizes
 *    - fetch the initial args and environment from the user space
 *    - allocate the memory for the new process
 *    - copy the initial stack from MM to the process
 *    - read in the text and data segments and copy to the process
 *    - take care of setuid and setgid bits
 *    - fix up 'mproc' table
 *    - tell kernel about EXEC
*
 * The entry points into this file are:
 *   do_exec:    perform the EXEC system call
*/
#include "COMMON\type.h"
#include "COMMON\com.h"
#include "COMMON\const.h"
#include "COMMON\stat.h"

#include "kernel\makra.h"
#include "fs\common\fsmes.h"
#include "systask\syst_pub.h"

#include "mm\tuoxhead.h"
#include "mm\mm_glob.h"
#include "mm\mmerror.h"
#include "mm\alloc.h"
#include "mm\mm.h"
#include "mm\exec.h"

#include "mm\mmtofs.h"
#include "test.cnf"

PRIVATE int get_stack(char*, Word*);
PRIVATE int get_path(char*, Word*);
PRIVATE int test_stack(char *stack, Word len, int *argc);
PRIVATE int read_header(int fdes, vir_bytes *text_bytes,
  vir_bytes *data_bytes, vir_bytes *bss_bytes, phys_bytes *tot_bytes,
  vir_clicks sc);
PRIVATE int new_mem(vir_bytes text_bytes, vir_bytes data_bytes,
  vir_bytes bss_bytes, vir_bytes stk_bytes, phys_bytes tot_bytes);
PRIVATE void patch_ptr(char stack[ARG_MAX], vir_bytes base);
PRIVATE int set_stack(char stack[ARG_MAX+MAIN_ARGS], vir_bytes *base,
  int argc, Word stk_len);
PRIVATE int load_seg(int fdes, int seg, vir_bytes seg_bytes);
PRIVATE int serve_setuid(char *filename);
PRIVATE int copy_mem_from(char *, char*, Word);
PRIVATE int copy_mem_to(char *, vir_bytes, Word);
PRIVATE int kernel_memmap(void);
PRIVATE int kernel_new_sp(Word sp);
PRIVATE int set_filename(char *filename);
int zero_mem(phys_clicks seg,vir_bytes fromOffset, vir_bytes len);

#if (TEST_PHASE<=MAX_PHASE_KERNEL_MM) && (TEST_PHASE!=0)
 extern Word k_code_segment, k_data_segment;
#endif

#define B2W(arr,i) (*((Word*)(&arr[(i)])))
#define B2D(arr,i) (*((DWord*)(&arr[(i)])))
#define W2B(arr,i) (*((Word*)(&arr[(i)])))


/*===========================================================================*
 *                              do_exec                                      *
 *===========================================================================*/
PUBLIC int do_exec()
{
/* Perform the execve(name, argv, envp) call.  The user library builds a
 * complete stack image, including pointers, args, environ, etc.  The stack
 * is copied to a buffer inside MM, and then to the new core image.
 */
  int err;
  Word stk_len;
  Word pth_len;
  int argc;
  int fdes;
  char temp_stack[ARG_MAX + MAIN_ARGS];
  char temp_path[PATH_MAX];
  char *filename;
  vir_bytes text_bytes, data_bytes, bss_bytes, sc, stack_base;
  phys_bytes tot_bytes;

/* Prevezmi stack pro execve() z user procesu do vnitrniho bufferu */
  err = get_stack(temp_stack + MAIN_ARGS, &stk_len);
  if (err) return err;
  err = test_stack(temp_stack + MAIN_ARGS, stk_len, &argc);
  if (err) return err;
/* Prevezmi path z user procesu do vnitrniho bufferu */
  err = get_path(temp_path, &pth_len);
  if (err) return err;

  stk_len+=MAIN_ARGS;

/* Over, jestli je zadany soubor spustitelny */
  filename = temp_path;
  err = test_exec(filename);
  if (err) return err;

/* Otevri soubor */
  err = open_file(filename, &fdes);
  if (err) return err;

/* Uloz nazev souboru do tabulky mproc */
  err = set_filename(filename);
  if (err) return err;

/* Precti header, extrahuj delky segmentu (text,data,bss) */

  sc  = (((stk_len + DEF_STACK_SIZE) + CLICK_SIZE) / CLICK_SIZE)*CLICK_SIZE;
   /* velikost stacku - zaokrouhli na paragrafy nahoru */

  err = read_header(fdes, &text_bytes, &data_bytes, &bss_bytes,
	  &tot_bytes, sc);
  if (err) return err;

/* Alokuj nove pametove oblasti, uvolni stare, informuj kernel */
  err = new_mem(text_bytes, data_bytes, bss_bytes, sc, tot_bytes);
  if (err) return err;

/* Ve vnitrnim bufferu patchuj stack, nakopiruj do nove image */
  stack_base = (vir_bytes)((tot_bytes - text_bytes) - stk_len);
  patch_ptr(temp_stack + MAIN_ARGS, stack_base);
  err = set_stack(temp_stack, &stack_base, argc, stk_len);
  if (err) return err;

/* Precti ze souboru obsah segmentu textu a dat */
  err = load_seg(fdes, T, text_bytes);
  if (err!=text_bytes) return (err);
  err = load_seg(fdes, D, data_bytes);
  if (err!=data_bytes) return (err);
  err = close_file(fdes);

/* Obsluz setuid/setgid bity */
  err = serve_setuid(filename);
  if (err) return err;

/* reset bitmapy chytanych signalu */
/*  mp->mp_ignore = 0;
    mp->mp_catch  = 0;
    mp->mp_func   = 0; */


/* Poslat NEW_SP kernelu */
  err = kernel_new_sp(stack_base);
  if (err) return err;

  return OK;
}



/*===========================================================================*
 *                              read_header                                  *
 *===========================================================================*/
PRIVATE int read_header(int fdes, vir_bytes *text_bytes,
  vir_bytes *data_bytes, vir_bytes *bss_bytes, phys_bytes *tot_bytes,
  vir_clicks sc)
/* int fdes;                 file descriptor for reading exec file
   int *ft;                place to return ft number
   vir_bytes *text_bytes;  place to return text size
   vir_bytes *data_bytes;  place to return initialized data size
   vir_bytes *bss_bytes;   place to return bss size
   phys_bytes *tot_bytes;  place to return total size
   vir_clicks sc;          stack size in clicks */
{

  /* Read the header and extract the text, data, bss and total sizes from it. */
  /* Read the header and check the magic number.  The standard TUOX header
   * consists of 8 longs:
   *    Bytes   0-3: Magic number
   *    Bytes   4-7: 0
   *    Bytes  8-11: size of text segments in bytes
   *    Bytes 12-15: size of initialized data segment in bytes
   *    Bytes 16-19: size of bss in bytes
   *    Bytes 20-24: total memory allocated to program (text, data + stack)

   * The header is followed directly by the text and data segments
   */

   int err;
   char header[EXE_HEADER_LEN];
   DWord Text, Data, Bss, Total;

   err = read_file_mm(fdes, header, EXE_HEADER_LEN, MM, D);
   if (err!=EXE_HEADER_LEN) if (err<0) return err; else return ENOEXEC;

   if (B2D(header, 0) != EXE_MAGIC_NUMBER)
     return ErrNotExecutable;
   Text = B2D(header, 8);
   if (Text > MAX_SEG) return ErrTextSizeTooBig;
   Data = B2D(header, 12);
   if (Data > MAX_SEG) return ErrDataSizeTooBig;
   Bss = B2D(header, 16);
   if (Bss > MAX_SEG) return ErrBssSizeTooBig;
   if (Data + Bss > MAX_SEG) return ErrDataBssSizeTooBig;
   Total = B2D(header, 20);
   if (Total > 2 * MAX_SEG) return ErrTotalSizeTooBig;
   if (Total - Text > MAX_SEG) return ErrTotalSizeTooBig;
   if (Total - (Text + Data + Bss) < sc) return ErrNotStackPlace;

   *text_bytes = (vir_bytes)Text; /* ((Text + 15) >> 4); */
   *data_bytes = (vir_bytes)Data;
   *bss_bytes  = (vir_bytes)Bss;
   *tot_bytes  = (phys_bytes)Total;

   return OK;
}


/*===========================================================================*
 *                              new_mem                                      *
 *===========================================================================*/
PRIVATE int new_mem(vir_bytes text_bytes, vir_bytes data_bytes,
  vir_bytes bss_bytes, vir_bytes stk_bytes, phys_bytes tot_bytes)
/*  vir_bytes text_bytes;   text segment size in bytes
    vir_bytes data_bytes;   size of initialized data in bytes
    vir_bytes bss_bytes;    size of bss in bytes
    vir_bytes stk_bytes;    size of initial stack segment in bytes
    phys_bytes tot_bytes;   total memory to allocate, including gap */
{
/* Allocate new memory and release the old memory.  Change the map and report
 * the new map to the kernel.  Zero the new core image's bss, gap and stack.
 */
   int err;
   Word total;
   phys_clicks text_seg, data_seg;
   struct mem_map *map;
   int theSame=0,i;

/* Zjisti, je-li pamet pro text,(bss+data),gap,stack: vsechny v celych
   poctech paragrafu, dohromady jeden spojity usek pameti
   Neni-li takovy: return (EAGAIN)
*/
   total = (Word)((tot_bytes + 15) >> 4);
   if (max_hole()<total)  return (EAGAIN);

   /* Uvolni stary blok pameti, trochu sloziteji (text mozna nekdo pouziva!) */
   map = mp->mp_seg;
/* Gap...*/

#if (TEST_PHASE<=MAX_PHASE_KERNEL_MM) && (TEST_PHASE!=0)
   if (map[D].mem_phys!=k_data_segment)
#endif
    free_mem(map[D].mem_phys,map[S].mem_vir+map[S].mem_len);


/* Text ? - prohledame zda nekdo (JINY!) nema stejny text */
//   for (i=0;i<NR_PROCS;i++)
   for (i=0;i<NR_PROCS+NR_TASKS+NR_SERVS;i++)
// NR_PROCS+NR_TASKS+NR_SERVS
     if (mproc[i].mp_flags&IN_USE)
       if (map[T].mem_phys==mproc[i].mp_seg[T].mem_phys)
	 if (mp!=&mproc[i])
         {
           theSame=1;
           break;
	 };
#if (TEST_PHASE<=MAX_PHASE_KERNEL_MM) && (TEST_PHASE!=0)
   if (map[T].mem_phys!=k_code_segment)
#endif
   if (!theSame) free_mem(map[T].mem_phys,map[T].mem_len);

   /* Alokuj novy usek, nastav mapu pameti */
   text_seg = alloc_mem(total);
   if (!text_seg) return ENOMEM;
   data_seg = text_seg + (text_bytes >> 4);

   map[T].mem_vir  = 0;
   map[T].mem_phys = text_seg;
   map[T].mem_len  = text_bytes / CLICK_SIZE;
   map[D].mem_vir  = 0;
   map[D].mem_phys = data_seg;
   map[D].mem_len  = (data_bytes + bss_bytes) / CLICK_SIZE;
   map[S].mem_vir  =
     (vir_clicks)((tot_bytes - text_bytes - stk_bytes));
   map[S].mem_phys = data_seg;
   map[S].mem_len  = stk_bytes / CLICK_SIZE;

   /* Vynuluj bss, gap, stack */
   err = zero_mem(data_seg, data_bytes,
      (vir_clicks)(tot_bytes - (text_bytes + data_bytes)));
   if (err) return err;

   /* Informuj kernel */
   err = kernel_memmap();
   if (err) return err;

   return OK;
}


/*===========================================================================*
 *                              patch_ptr                                    *
 *===========================================================================*/
PRIVATE void patch_ptr(char stack[ARG_MAX], vir_bytes base)
/* char stack[ARG_MAX];    pointer to stack image within MM
   vir_bytes base;         virtual address of stack base inside user */
{
/* When doing an exec(name, argv, envp) call, the user builds up a stack
 * image with arg and env pointers relative to the start of the stack.  Now
 * these pointers must be relocated, since the stack is not positioned at
 * address 0 in the user's address space.
 */
   int i;
   Word *a = (Word*) stack;
   for (i = 0; i < 2; i++)
     {
       while (*a) *a++ += base;
       a++;
     }
}


/*===========================================================================*
 *                              load_seg                                     *
 *===========================================================================*/
PRIVATE int load_seg(int fdes, int seg, vir_bytes seg_bytes)
/* int fdes;              file descriptor to read from
   int seg;               T or D
   vir_bytes seg_bytes;   how big is the segment         */
{
/* Read in text or data from the exec file and copy to the new core image. */
  return read_file_mm(fdes,0,seg_bytes,who,seg);
}


/*===========================================================================*
 *                              get_stack                                    *
 *===========================================================================*/
/*  prekopirovani stacku do pameti MM  */
PRIVATE int get_stack(char dest[ARG_MAX], Word *stk_len)
{
  int err;
  char *sour;
  sour = mm_in.MM_EXECSTACK;
  *stk_len = mm_in.MM_EXECSIZE;
  err = copy_mem_from(sour, dest, *stk_len);
  return err;
}
/*===========================================================================*
 *                              get_path                                     *
 *===========================================================================*/
/*  prekopirovani cesty do pameti MM  */
PRIVATE int get_path(char dest[PATH_MAX], Word *pth_len)
{
  int err;
  char *sour;
  sour = mm_in.MM_EXECFNAME;
  *pth_len = mm_in.MM_PATHSIZE;
  err = copy_mem_from(sour, dest, *pth_len);
  return err;
}


/* otestovani predaneho stacku */
PRIVATE int test_stack(char stack[ARG_MAX], Word len, int *argc)
{
  Word *a, b, i, j;
  a = (Word*)stack;
  /* na zacatku stacku je pole near pointru na argumenty - */
  /* spocitej kolik jich je (jsou ukonceny 0)              */
  i = 0;
  while (i < len)
    {
      if (*a++ == 0) break;
      i += 2;
    }
  *argc = i / 2;
  /* musi byt alespon jeden argument - nazev souboru */
  if (*argc < 1) return ErrMissingFilename;
  /* nasleduje pole near pointru na environment promenne - projdi je */
  while (i < len)
    {
      i += 2;
      if (*a++ == 0) break;
    }
  /* pokud uz ses dostal na konec stacku je to chyba */
  if (i >= len) return ErrBadStack;
  /* projdi vsechny polozky - jak argumenty, tak env. promenne   */
  /* kazdy retezec musi lezet mezi koncem pole env. promennych a */
  /* koncem stacku a musi byt ukoncen 0                          */
  a = (Word*)stack;
  for (j = 0; j < 2; j++)
    while ((b = *a++) != 0)
      {
	if (b < i || b >= len) return ErrBadStack;
	while (stack[b])
	  if (b++ == len) return ErrBadStack;
      }
  return OK;
}


/* nastaveni noveho stacku, na zacatek pridan pocet argumentu, ukazatel    */
/* na zacatek pole pointru na argumenty a ukazatel na zacatek pole pointru */
/* na env. promenne                                                        */
PRIVATE int set_stack(char stack[ARG_MAX+MAIN_ARGS], vir_bytes *base,
  int argc, Word stk_len)
{
  int err;
  W2B(stack,0) = argc;
  W2B(stack,2) = *base;
  W2B(stack,4) = *base + 2 * argc + 2;
  stk_len += MAIN_ARGS;
  *base -= MAIN_ARGS;
  err = copy_mem_to(stack, *base, stk_len);
  return err;
}


/* obslouzeni setuid a setgid bitu */
PRIVATE int serve_setuid(char *filename)
{
  int err;
  int bit;
  gid_t owner_g;
  uid_t owner;
  /* efektivni uid = skutecne uid */
  mp->mp_effuid = mp->mp_realuid;
  /* zjisti, zda je u souboru nastaven setuid bit           */
  /* pokud ano, nastav efektivni uid podle vlasnika souboru */
  err = get_setuid(filename, &bit);
  if (err) return err;
  if (bit)
    {
      err = get_file_owner(filename, &owner);
      if (err) return err;
      mp->mp_effuid = owner;
    }
  /* efektivni gid = skutecne gid */
  mp->mp_effgid = mp->mp_realgid;
  /* zjisti, zda je u souboru nastaven setgid bit           */
  /* pokud ano, nastav efektivni gid podle vlasnika souboru */
  err = get_setgid(filename, &bit);
  if (err) return err;
  if (bit)
    {
      err = get_file_owner_g(filename, &owner_g);
      if (err) return err;
      mp->mp_effgid = owner_g;
    }
  return OK;
}


/* prekopirovani casti pameti z procesu do MM */
PRIVATE int copy_mem_from(char *sour, char *dest, Word len)
{
  mm_out.m_type  = SYS_COPY;
  mm_out.SRC_PROC = who;
  mm_out.SRC_SEG = D;
  mm_out.SRC_VIR = (Word) sour;

  mm_out.DST_PROC = MM;
  mm_out.DST_SEG = D;
  mm_out.DST_VIR = (Word)dest;
  mm_out.BYTE_CNT = len;
  send_recv(SYSTASK, mm_out);
  return mm_out.RET_ERROR_CODE;
}


/* prekopirovani casti pameti z MM do procesu */
PRIVATE int copy_mem_to(char *sour, vir_bytes dest, Word len)
{
  mm_out.m_type  = SYS_COPY;    /* kopirovani bloku pameti */
  mm_out.SRC_PROC = MM;
  mm_out.SRC_SEG = D;
  mm_out.SRC_VIR = (Word) sour;

  mm_out.DST_PROC = who;
  mm_out.DST_SEG = D;
  mm_out.DST_VIR = (Word)dest;
  mm_out.BYTE_CNT = len;
  send_recv(SYSTASK, mm_out);   /* posli zpravu */
  return mm_out.RET_ERROR_CODE;
}


/* informuj kernel o nove nastavene mape pameti */
PRIVATE int kernel_memmap()
{
  mm_out.m_type  = SYS_NEWMAP;  /* nova mapa pameti  */
  mm_out.PROC  = who;           /* cislo procesu     */
  mm_out.MEM_PTR = (char*)(mp->mp_seg);   /* nova mapa pameti  */
  send_recv(SYSTASK, mm_out);  /* posli zpravu      */
  return OK;
}


/* provedeni execu, nastaveni nove hodnoty sp */
PRIVATE int kernel_new_sp(Word sp)
{
  mm_out.m_type  = SYS_EXEC;     /* proved exec, nastav sp */
  mm_out.PROC = who;             /* cislo procesu          */
  mm_out.NEW_SP = sp;            /* nova hodnota sp        */
  send_recv(SYSTASK, mm_out);    /* posli zpravu           */
  return OK;
}


/* nastaveni nazvu souboru v tabulce mproc */
PRIVATE int set_filename(char *filename)
{
  char *p, *fn;

  /* najdi konec retezce */
  p = filename;
  while (*p) p++;

  while (p != filename)  /* pokud jsi dosel zpet az na zacatek retezce */
    {                    /*  skonci                                    */
       if (*p == '/')  {  /* az naleznes '/', skonci */
	  p++;
	  break;
       }
       p--;
    }
  /* prekopiruj nazev souboru do tabulky mproc */
  fn = mp->mp_filename;
  do {
    *fn++ = *p;
  } while (*p++);

  return OK;
}

/* Vynuluje bss, gap, stack */
int zero_mem(phys_clicks seg,vir_bytes fromOffset, vir_bytes len) {

 return 0;
}
