/*
 ****************************************************************************
 * soubor: forkexit.c                                                       *
 * autor:  Marcel Kruzel, Zdenek Sawa, Milan Gvozdik                        *
 * ladici: Rojicek Radim, Sela Jaroslav                                     *
 ****************************************************************************
*/

/*
 * This file deals with creating processes (via FORK) and deleting them (via
 * EXIT/WAIT).  When a process forks, a new slot in the 'mproc' table is
 * allocated for it, and a copy of the parent's core image is made for the
 * child.  Then the kernel and file system are informed.  A process is removed
 * from the 'mproc' table when two events have occurred: (1) it has exited or
 * been killed by a signal, and (2) the parent has done a WAIT.  If the process
 * exits first, it continues to occupy a slot until the parent does a WAIT.
 *
 * The entry points into this file are:
 *   do_fork:    perform the FORK system call
 *   do_mm_exit: perform the EXIT system call (by calling mm_exit())
 *   mm_exit:    actually do the exiting
 *   do_wait:    perform the WAITPID or WAIT system call
 */


#include "COMMON\callnr.h"
#include "COMMON\type.h"
#include "COMMON\com.h"
#include "COMMON\const.h"
#include "test\trap.h"

#include "kernel\makra.h"
#include "systask\syst_pub.h"

#include "mm\forkexit.h"
#include "mm\alloc.h"
#include "mm\mm.h"
#include "mm\mmerror.h"
#include "mm\mm_glob.h"
#include "mm\reply.h"

#include "fs\common\fsmes.h"
#include "mm\mmtofs.h"
#include "test.cnf"


extern int memcpy(char *,char *,int); /* mmemcpy.c */



PRIVATE int wait(int parent,pid_t waitFor,int options);
PRIVATE int fork(int parent,struct mproc **rmc);
PRIVATE void cleanup(int child);
PRIVATE pid_t next_pid = INIT_PID+1;    /* next pid to be assigned */

/*===========================================================================*
 *                              do_fork                                      *
 *===========================================================================*/

PUBLIC int do_fork()
{
  /* Obsluha zpravy FORK */
  struct mproc *child;
  int err=fork(who,&child);
  if (err==OK) mm_out.MM_FORKEDCHILD=child->mp_pid;
  return err;
};


PRIVATE int fork(int parent,   /* index to parent */
		 struct mproc **rmc /* pointer to child */)
{
/* The process pointed to by 'mp' has forked.  Create a child process. */
  vir_clicks gapSize;
  phys_clicks dataMem;
  int posit;
  struct mproc *pom,*rmp=&mproc[parent];
  /* Jsou-li vsechny sloty obsazeny, return(EAGAIN);
     Zbyva-li mene nez LAST_FEW slotu a efektivni uid neni 0: return EAGAIN;
  */
  if ((procs_in_use==NR_PROCS)||(NR_PROCS-procs_in_use<=LAST_FEW)&&(rmp->mp_effuid!=0)) return EAGAIN;

  /* Velikost gapu: Soucet virualni adresy stacku + velikost stacku.
     Zjisti, kolik pameti se bude kopirovat. Kopiruje se jen data
     a stack (+prostor v GAP), text segment je sdileny.
  */
  gapSize=(rmp->mp_seg[S].mem_vir+CLICK_SIZE-1)/CLICK_SIZE+rmp->mp_seg[S].mem_len;
  /* Zkus naalokovat pamet, nepovede-li se: return(ENOMEM); */
  dataMem=alloc_mem(gapSize);
  if (dataMem==0) return ENOMEM;

  posit=0;
  while (mproc[posit].mp_flags&IN_USE) posit++;
  pom=&mproc[posit];
  procs_in_use++;

  /* Zkopiruj informace ze slotu rodice */
  memcpy((char *)pom,(char*)rmp,sizeof(struct mproc));

  /*Nastav v nem memory map, parent; mp_exitstatus=mp_sigstatus=0;*/
  pom->mp_seg[D].mem_phys=dataMem;
  pom->mp_seg[S].mem_phys=dataMem;
  /* segment T bude sdileny */

  pom->mp_exitstatus=pom->mp_sigstatus=0;
  /*Nastaveni parenta */
  pom->mp_parent=parent; /* Index v tabulce mproc */
  /* Vymysli PID pro novy proces: bez overflow (zatim -- tolik procesu se stejne nepodari rozjet :-) */
  pom->mp_pid=next_pid++;
/*
  Oznam uspesny fork kernelu (SYS_NEWMAP) a FS (aby mohli vytvorit slot ve svych
  castech tabulky procesu)
*/

/* SYS_FORK */
  mm_out.PARENT=parent;
  mm_out.CHILD=posit;
  mm_out.PID=pom->mp_pid;
  mm_out.m_type=SYS_FORK;
  send_recv(SYSTASK,mm_out);
  if (mm_out.RET_ERROR_CODE!=0) trap(TRAP_FORK_SFORKERR);
/* SYS_NEWMAP */
  mm_out.PROC=posit;
  mm_out.MEM_PTR=(char *)&(mproc[posit].mp_seg);
  mm_out.m_type=SYS_NEWMAP;
  send_recv(SYSTASK,mm_out);
  if (mm_out.RET_ERROR_CODE!=0) trap(TRAP_FORK_SNEWMAPERR);
/* SYS_COPY */
  mm_out.SRC_SEG=D;
  mm_out.SRC_PROC=parent;
  mm_out.SRC_VIR=0;
  mm_out.DST_SEG=D;
  mm_out.DST_PROC=posit;
  mm_out.DST_VIR=0;
  mm_out.BYTE_CNT=gapSize*16;  /* v bajtech */
  mm_out.m_type=SYS_COPY;
  send_recv(SYSTASK,mm_out);
  if (mm_out.RET_ERROR_CODE!=0) trap(TRAP_FORK_COPYERR);

/* NEWSLOT TO FS */

#if ((TEST_PHASE>MAX_PHASE_KERNEL_MM) || (TEST_PHASE==0) \
      || defined(LOAD_FS_MODULE))
  mm_out.fs_parent=parent;
  mm_out.fs_child=posit;
  mm_out.m_type=C_FORK;
  send_recv(FS,mm_out);
#endif

  mm_out.MM_FORKEDCHILD=0;
   /*  Odpovez detskemu procesu (tim se probudi)   */
  reply(posit,OK);
  /* .. rodicovskemu procesu odpovi hlavni smycka. */
  *rmc=pom;
  return OK;  /* Vrat OK ... */
}


/*===========================================================================*
 *                              do_mm_exit                                   *
 *===========================================================================*/
PUBLIC int do_mm_exit()
{
/* Perform the exit(status) system call. The real work is done by mm_exit(),
 * which is also called when a process is killed by a signal.
 */
  mm_exit(who, mm_in.MM_EXITSTATUS);  /* mp je pripravena v main */
  dont_reply = TRUE;                 /* don't reply to newly terminated process */
  return OK;
}


/*===========================================================================*
 *                              mm_exit                                      *
 *===========================================================================*/
PUBLIC void mm_exit(register int child, char exit_status)
 /* pointer to the process to be terminated +
    the process' exit status (for parent)   */
{
  struct mproc *rmp=&mproc[child];
/* A process is done.  Release most of the process' possessions.  If its
 * parent is waiting, release the rest, else hang.
 */
 /* exitstatus do mproc[] */
  rmp->mp_exitstatus=exit_status;
/* SYS_XIT */
  mm_out.PARENT=rmp->mp_parent;
  mm_out.PROC=child;
  mm_out.m_type=SYS_XIT;
  send_recv(SYSTASK,mm_out);

  /* informuj FS */

 #if ((TEST_PHASE>MAX_PHASE_KERNEL_MM) || (TEST_PHASE==0) \
      || defined(LOAD_FS_MODULE))

  mm_out.fs_proc=child;
  mm_out.m_type=C_EXIT;
  send_recv(FS,mm_out);
 #endif

  if ((mproc[rmp->mp_parent].mp_flags & WAITING)||
      (mproc[rmp->mp_parent].mp_flags & WAITINGFOR)&&(mproc[rmp->mp_parent].mp_waitfor==rmp->mp_pid)||
      (mproc[rmp->mp_parent].mp_flags & WAITINGGROUP)&&(mproc[rmp->mp_parent].mp_waitfor==rmp->mp_procgrp)
     ) cleanup(child);

  else rmp->mp_flags |= HANGING; /* Zombie until parent waits */
  /* Zrus pripadne alarmy (ALARM_ON z mp_flags). */
  rmp->mp_flags&=~ALARM_ON;
}


/*===========================================================================*
 *                              cleanup                                      *
 *===========================================================================*/
PRIVATE void cleanup(int child /* tells which process is exiting */)
{
  int theSame=0,               /* nekdo jiny ma stejny textovy segment */
      processes=procs_in_use,  /* kolik polozek mprocu obsazeno */
      pom=0;     /* Pomocny index do tabulky mproc */
  phys_clicks aText=mproc[child].mp_seg[T].mem_phys; /* Adresa text segmentu */
  struct mproc *cmp=&mproc[child];
/* Rusime tento proces */
  cmp->mp_seg[T].mem_phys=0;
/* Probuzeni rodice, jakkoliv cekal. */
  mproc[cmp->mp_parent].mp_flags&=~(WAITINGGROUP|WAITING|WAITINGFOR);
  mm_out.MM_EXITVALUE=(unsigned int)cmp->mp_exitstatus+(cmp->mp_sigstatus<<8);
  mm_out.MM_EXITEDPID=cmp->mp_pid;
  reply(cmp->mp_parent,OK);
/* Oznameni FS - kernelu uz se to reklo pri volani exit() */

#if ((TEST_PHASE>MAX_PHASE_KERNEL_MM) || (TEST_PHASE==0) \
      || defined(LOAD_FS_MODULE))

  mm_out.fs_proc=child;
  mm_out.m_type=C_EXIT;
  send_recv(FS,mm_out);
#endif

/* dealokuj pamet zabranou procesem. */
/* Gap: */
  free_mem(cmp->mp_seg[D].mem_phys,cmp->mp_seg[S].mem_vir+cmp->mp_seg[S].mem_len);
/* Jestlize vsechny ostatni procesy maji jiny TEXT: zrus CHILD::TEXT.*/
/* Projit vsechny procesy, zda nemaji stejny text segment. */
 while (processes>0)           /* Pres vsechny aktivni procesy */
 {
   if (mproc[pom].mp_flags&IN_USE) 
   {
     if (aText==mproc[pom].mp_seg[T].mem_phys) theSame=1; /* Jiny proces ma stejny TEXT,
                                              jako child (ale neni to child (ten ma==0))*/ 
     if (mproc[pom].mp_parent==child)   /* child je rodicem pomu */
     {
       mproc[pom].mp_parent=INIT_PID;       /* Init adoptuje kazde dite procesu child */
       /* Nasledujici if bude platny nejvyse jednou behem celeho cyklu a ve vnoreni uz vubec ne,
          nebot Init cekat muze jen na jednoho a wait se mu po probuzeni nepodari 
          uskutecnit (MM zaneprazdnen). Nevadi nam, ze procs_in_use bude po navratu 
          mensi (bude zrusena probrana v cyklu polozka pom). Pokud mel pom stejny 
          textovy segment (a nikdo jiny) jako child, a init je waiting pak ho zrusi vnitrni 
          cleanup, ne my. Nad touto skutecnosti jsem premyslel asi 2 hodiny, 
          tak tam nehledejte chyby prosim! Dale se predpoklada, ze init ceka jednoduse 
	  (WAITING a ne WAITINGGROUP ci podobne). Tento predpoklad je si myslim opravneny.
       */
       if ((mproc[INIT_PID].mp_flags & WAITING) && (mproc[pom].mp_flags & HANGING)) cleanup(pom);         
     };
     processes--;
   };
   pom++;
 };
 /* nebyl-li nikde text stejny */
  if (!theSame) free_mem(aText,cmp->mp_seg[T].mem_len);
  procs_in_use--; 
  cmp->mp_flags=~IN_USE; /*uklid i ostatni flagy */
/* Vyse udelano: Pokud mel ukoncovany proces potomky (CH), preved jejich rodicovstvi na Init: */
/* Vyse udelano: Pokud je mproc[Init].mp_flags WAITING a nektery z CH je HANGING: cleanup*/
}


/*===========================================================================*
 *                              wait                                         *
 *===========================================================================*/
PUBLIC int do_wait()
{
  int err;      /* Pokud neco nehraje, odpovez; jinak nic. */
  if ((err=wait(who,(pid_t)mm_in.MM_WAITFOR,mm_in.MM_WAITOPT))==OK) dont_reply=TRUE;
  return err;
};

PRIVATE int wait(int parent,pid_t waitFor,int options)
/* V parent je index do mproc volajiciho procesu */
{
/*  .if (waitFor==-1) ANY; 
    .if (waitFor==0)  The Same group ID as Parent
    .if (waitFor>0) Child with pid_t
    .if (waitFor<-1) Child with group==|pid_t|
*/


  int pom=INIT,        /* Iterace tabulkou mproc */
      found=FALSE,  /* Vlajka zda jsme nalezl dite */
      processes=procs_in_use;
  struct mproc *ppom;  /* Odpovida pom */
  /* Projed tabulku procesu, najdi potomky procesu volajiciho wait.
     Dle waitFor a options urci co dal?
  */
  while (processes>0)                     /* Pres vsechny existujici procesy */
  {
    if ((ppom=&mproc[pom])->mp_flags&IN_USE)
    {
      processes--;
      if (ppom->mp_parent==parent)           /* Nasel jsem dite */
      {
	if ((waitFor==ppom->mp_pid)&&(waitFor>0))  /* Waits for exactly this one. */
	{
	  if (ppom->mp_flags&HANGING)              /* Already finished */
	  {
	    cleanup(pom);
            return OK;
          } else
          {
            if (options&WNOHANG) 
              return ErrChildStillAlive;   /* Odpovime, ze dite jeste neskoncilo */              
            mproc[parent].mp_flags|=WAITINGFOR;    /* Cekam na konkretni dite */
            mproc[parent].mp_waitfor=waitFor;            
            return OK;
          };
        } else
	if ((waitFor<-1)&&(-waitFor==ppom->mp_procgrp))    /* Waits for child in this group. */
        {
	  if (ppom->mp_flags&HANGING)
          {
            cleanup(pom);
            return OK;
          } else
          {
            mproc[parent].mp_flags|=WAITINGGROUP;
            mproc[parent].mp_waitfor=-waitFor;
            found=1;                              /* Snad najdu jiny: hanging */
          };
        } else 
        if (waitFor==-1)                  /* ANY Child */
        {          
	  if (ppom->mp_flags&HANGING)
          {
            cleanup(pom);
            return OK;
          } else 
	  {
            mproc[parent].mp_flags|=WAITING; /* Pokud jiny child hanging, flag se znuluje v cleanupu */
            found=1;
          };
        } else
        if (waitFor==0)                  /* Any child with the same group ID */
          if (ppom->mp_procgrp==mproc[parent].mp_procgrp)
            if (ppom->mp_flags&HANGING)
	    {
              cleanup(pom);
              return OK;
            } else 
            {
              mproc[parent].mp_flags|=WAITINGGROUP;
	      mproc[parent].mp_waitfor=mproc[parent].mp_procgrp;
              found=1;
            };
      };        
    };
    pom++;
  }

  if (!found) return ECHILD;   /* No child process exists */
  else if (options&WNOHANG) 
    {
      mproc[parent].mp_flags&=~(WAITINGGROUP|WAITING|WAITINGFOR);
      return ErrChildStillAlive;
    };
  return OK;
};