/*******************************************************************
*                       SYSTASK.H
*	Soucast projektu TUOX
*	Autori: Jiri Horacek		 
*		Alan Kornuta
*		David Stupnanek
*	Upravil: Petr Grygarek
*/

/* Tento task slouzi jako interface mezi MM a kernelem, resp. mezi FS
   a kernelem. Implementovane sluzby resi zejmena ty casti systemovych
   volani, ktere operuji nad tabulkou proc[]. MM a FS k teto tabulce
   nemohou pristupovat primo, jelikoz neni soucasti jejich adresoveho
   prostoru. Mimo to je zde obsazena sluzba pro kopii bloku pameti
   mezi adresovymi prostory dvou procesu a sluzba pro zahajeni vykonavani
   obsluzne rutiny signalu.

   Podporovane zpravy:

     SYS_COPY    Kopie bloku pameti mezi adresovymi prostory procesu

     SYS_FORK    Informuje kernel o rozstepeni procesu
     SYS_NEWMAP  Nastavi novou pametovou mapu procesu (iniciuje MM)
     SYS_EXEC    Po vymene image nastavi PC na zacatek programu a nastavi SP
     SYS_XIT     Informuje kernel o ukonceni procesu
     SYS_GETSP   Precteni hodnoty registru SP urceneho procesu (pro brk())
     SYS_TIMES   Zjisteni casu spotrebovanych procesem
     SYS_SIG     Zaslani signalu procesu

     Dale SYSTASK obsahuje nekolik rutin, ktere mohou byt volany
     ostatnimi tasky (a jadrem), ktere budou slinkovany spolecne
     s nim:

     cause_sig(): naplanovani doruceni signalu procesu
     inform():    informace MM, aby vyridil nedorucene signaly
     umap():      vypocet fyzicke adresy z virtualni
     sys_copy()	  totez jako zprava SYS_COPY
*/


/*
	  Makra pro symbolicke pojmenovani clenu jednotlivych zprav:
	  ----------------------------------------------------------
*/

#include "systask\syst_pub.h"

/*===========================================================================*
				cause_sig
		 Zaslani signalu procesu nekterym taskem
   args:
	int proc_nr		- cislo procesu, kteremu se posle signal
	int sig_nr		- cislo zasilaneho signalu

   Vola se, pokud chce task zaslat nejakemu procesu signal. Napr.
   - TTY zasila SIGINT pri stisku DEL,
   - CLOCK zasila SIGALRM pri vyprseni nastaveneho casoveho intervalu.

   Signaly ve skutecnosti obsluhuje MM (zadost o obsluhu zpravou KSIG).
   Tasky tuto zpravu nezasilaji primo, protoze MM nemusi byt prave
   schopen zpravu prijmout. Rutina cause_sig() nastavi bit v bitmape
   aktivovanych signalu - proc[].p_pending a vola inform(), ktery se pokousi
   zpravu MM dorucit. Nepovede-li se to okamzite, ucini se tak pozdeji
   z kernelu, kdyz MM pozada o zpravu a zadna jina zprava pro nej
   nebude k dispozici.
 *===========================================================================*/
void cause_sig(int proc_nr, int sig_nr);


/*===========================================================================
				inform
 Pokus o zaslani zpravy KSIG memory manageru pro kazdy proces,
 kteremu ma byt dorucen signal (proc[].p_pending!=0).
 Vola se z cause_sig() a pak z jadra, vzdy, kdyz  MM zavola receive(ANY)
 a zadna jina zprava pro nej neni (tj. MM nema nic na praci).
 ===========================================================================*/
void inform(void);


/*===========================================================================
				umap
	mapovani adresy z adr. prostoru procesu na linearni
   args:
	int proc_nr		- cislo proscesu v proc[]
	int seg			- segment (T, D)
	vir_bytes vir_addr	- virtualni adresa v ramci segmentu
	vir_bytes count		- pocet kopirovanych bytu (test jestli
				  jsme porad uvnitr segmentu)
				  0=netestovat	
   ret:				
	long      		- spoctena fyzicka (linearni) adresa
				  nebo chybovy kod:
				       ESRCH - proces je neplatny
				       ENOMEM - preteceni segmentu (od count)
 ============================================================================*/
 long umap(int proc_nr, int seg, vir_bytes vir_addr, vir_bytes count);


/*===========================================================================*
                             sys_copy
 	 Kopirovani meze virtualnimi adresovymi prostory.
   args:
 	src_proc, dst_proc  	- pozice zdrojoveho/ciloveho procesu v proc[]
        src_seg, dst_seg	- segment (T,D)
 	src_addr, dst_addr	- virtualni adresa v ramci segmentu
        count			- pocet kopirovanych bytu
   ret:
 
          OK
          ESRCH - proces je neplatny
          EFAULT - chybna virtualni adresa
          ENOMEM - blok presahuje (src/dest) segment
 *===========================================================================*/

int sys_copy(procnr_t src_proc, Byte src_seg /* T/D */ , vir_bytes src_addr,
	      procnr_t dst_proc, Byte dst_seg /* T/D */, vir_bytes dst_addr,
	      Word count);

/*===========================================================================
				sys_task
		Vstupni bod. Inicializace, hlavni smycka
 ===========================================================================*/
void sys_task(void);
